
import torch
import os
import numpy as np
import torch.nn as nn
import torch.optim as optim
from tqdm import tqdm
from .metric import evaluate
from .preprocess import preprocess_data

from .._model import MLmodel
from . import param_default

os.environ["CUDA_VISIBLE_DEVICES"] = "0"
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

param_default_dict = {
    'network': 'CNN',
    'net_type': 'gru',
    'layers': 1,
    'seq_len': 256,
    'input_size': 1,
    'hidden_size': 100,
    'rep_size': 20,
    'batch_size': 64,
    'max_epoch': 5,
    'lr': 0.01,
    'lambda': 1,
    'out_path': './output'
}


class Generator(nn.Module):
    def __init__(self, nc=25, nz=100, seq_len=128, device=None):
        super(Generator, self).__init__()
        self.device = device
        self.nc = nc
        self.nz = nz
        self.ndf = 32

        encoder_modules = []
        encoder_modules.append(nn.Conv1d(self.nc, self.ndf, 4, 2, 1, bias=False))
        encoder_modules.append(nn.LeakyReLU(0.2, inplace=True))

        cur_len = seq_len//2
        i = 1
        while cur_len > 5:
            encoder_modules.append(nn.Conv1d(self.ndf*i, self.ndf * i*2, 4, 2, 1, bias=False))
            encoder_modules.append(nn.BatchNorm1d(self.ndf * i*2))
            encoder_modules.append(nn.LeakyReLU(0.2, inplace=True))
            i = i*2
            cur_len = cur_len//2

        encoder_modules.append(nn.Conv1d(self.ndf * i, self.nz, cur_len, 1, 0, bias=False))

        self.encoder = nn.Sequential(*encoder_modules)

        decoder_modules = []
        decoder_modules.append(nn.ConvTranspose1d(self.nz, self.ndf * i, 4, 1, 0, bias=False))
        decoder_modules.append(nn.BatchNorm1d(self.ndf * i))
        decoder_modules.append(nn.ReLU(True))
        i = i//2

        while i >= 1:
            decoder_modules.append(nn.ConvTranspose1d(self.ndf*(i*2), self.ndf * i, 4, 2, 1, bias=False))
            decoder_modules.append(nn.BatchNorm1d(self.ndf * i))
            decoder_modules.append(nn.ReLU(True))
            i = i//2
        decoder_modules.append(nn.ConvTranspose1d(self.ndf, self.nc, 4, 2, 1, bias=False))
        decoder_modules.append(nn.Tanh())

        self.decoder = nn.Sequential(*decoder_modules)

        print(encoder_modules)
        print(decoder_modules)

    def forward(self, input):
        input = torch.transpose(input, 1, 2)
        z = self.encoder(input)
        out = self.decoder(z)
        out = torch.transpose(out, 1, 2)
        return out, z


class Discriminator(nn.Module):
    def __init__(self, nc=25, seq_len=128, device=None):
        super(Discriminator, self).__init__()
        self.device = device
        self.nc = nc
        self.ndf = 32

        modules = []
        modules.append(nn.Conv1d(self.nc, self.ndf, 4, 2, 1, bias=False))
        modules.append(nn.LeakyReLU(0.2, inplace=True))

        cur_len = seq_len // 2
        i = 1
        while cur_len > 5:
            modules.append(nn.Conv1d(self.ndf * i, self.ndf * i * 2, 4, 2, 1, bias=False))
            modules.append(nn.BatchNorm1d(self.ndf * i*2))
            modules.append(nn.LeakyReLU(0.2, inplace=True))
            i = i * 2
            cur_len = cur_len // 2

        self.features = nn.Sequential(*modules)

        self.classifer = nn.Sequential(
            nn.Conv1d(self.ndf * i, 1, cur_len, 1, 0, bias=False),
            # state size. 1
            nn.Sigmoid()
        )

    def forward(self, input):
        input = torch.transpose(input, 1, 2)

        features = self.features(input)
        out = self.classifer(features).view(-1, 1).squeeze(1)

        return out, features


def weights_init(mod):
    """
    Custom weights initialization called on netG, netD and netE
    :param m:
    :return:
    """
    classname = mod.__class__.__name__
    if classname.find('Conv') != -1:
        # mod.weight.data.normal_(0.0, 0.02)
        nn.init.xavier_normal_(mod.weight.data)
        # nn.init.kaiming_uniform_(mod.weight.data)

    elif classname.find('BatchNorm') != -1:
        mod.weight.data.normal_(1.0, 0.02)
        mod.bias.data.fill_(0)
    elif classname.find('Linear') != -1:
        torch.nn.init.xavier_uniform(mod.weight)
        mod.bias.data.fill_(0.01)


class BeatGAN(MLmodel):
    def __init__(self, dataloader):
        super(BeatGAN, self).__init__(dataloader)
        self.dataloader = None
        self.device = None
        self.lamda_value = None

        self.generator = None
        self.discriminator = None

        self.mse = None
        self.bce = None

        self.optimizerG = None
        self.optimizerD = None

        self.out_dir = None
        self.iteration = None

        self.fix_input = None

    def train(self, data=None):
        self.generator.apply(weights_init)
        self.discriminator.apply(weights_init)
        self.generator.train()
        self.discriminator.train()

        best_auc = 0
        best_auc_epo = 0
        best_f1 = 0
        best_f1_epo = 0

        for epoch in tqdm(range(self.max_epoch)):
            self.train_epoch()
            # val_auc = self.val_epoch()
            # test_auc, test_th, test_f1 = self.test(intrain=True)

            self.save_cur_model(self.out_dir, "cur_w.pth")

            # if val_auc > best_val_auc:
            #     best_val_auc = val_auc
            #     best_val_auc_epo = epoch
            #     self.save_cur_model(self.out_dir, "beat_val_auc_w.pth")

            # if test_auc > best_auc:
            #     best_auc = test_auc
            #     best_auc_epo = epoch
            #     self.save_cur_model(self.out_dir, "beat_auc_w.pth")
            #
            # if test_f1 > best_f1:
            #     best_f1 = test_f1
            #     best_f1_epo = epoch
            #     self.save_cur_model(self.out_dir, "beat_f1_w.pth")

            # print(
            #     "[{}]test auc:{:.4f} th:{:.4f} f1:{:.4f} \n".format(epoch, test_auc, test_th, test_f1))
            # print("best auc:{} in [{}]    best f1:{}  in [{}]".format(best_auc, best_auc_epo, best_f1, best_f1_epo))
            # print("best val auc:{} in [{}]".format(best_val_auc, best_val_auc_epo))

    def train_epoch(self):
        self.generator.train()
        self.discriminator.train()

        for i, data in enumerate(self.dataloader):
            self.iteration += 1

            data_X, y = data

            data_X = data_X.to(self.device)

            if self.fix_input is None:
                self.fix_input = data_X

            # train discriminator
            self.discriminator.zero_grad()

            # train with real

            d_real_prob, _ = self.discriminator(data_X)

            # train with fake
            fake_data, enc_z = self.generator(data_X)
            d_fake_prob, _ = self.discriminator(fake_data)

            loss_d_real = self.bce(d_real_prob, torch.full((d_real_prob.size(0),), 1, dtype=torch.float, device=self.device))
            loss_d_fake = self.bce(d_fake_prob, torch.full((d_fake_prob.size(0),), 0, dtype=torch.float, device=self.device))

            loss_d = loss_d_real + loss_d_fake
            loss_d.backward()
            self.optimizerD.step()

            # train generator
            self.generator.zero_grad()

            fake_data, _ = self.generator(data_X)
            d_real_prob, d_real_feat = self.discriminator(data_X)
            d_fake_prob, d_fake_feat = self.discriminator(fake_data)

            loss_g_rec = self.mse(data_X, fake_data)
            loss_g_adv = self.mse(d_real_feat, d_fake_feat)

            loss_g = loss_g_rec + loss_g_adv * self.lamda_value
            loss_g.backward()

            self.optimizerG.step()

            if (i % 100) == 0:
                print("{}:loss_g(rec/adv):{}/{},loss_d(real/fake):{}/{}".format(i, loss_g_rec,
                                                                                loss_g_adv * self.lamda_value,
                                                                                loss_d_real, loss_d_fake))
                # fake_data, z = self.generator(self.fix_input)
                # show_num = 8
                # img_tensor = multi_ts2image(fake_data.detach().cpu().numpy()[:show_num],
                #                             self.fix_input.cpu().numpy()[:show_num])
                # for img_idx in range(show_num):
                #     self.writer.add_image("image/rec" + str(img_idx), img_tensor[img_idx], global_step=self.iteration,
                #                           dataformats='HWC')

            if loss_d.item() < 5e-6:
                self.discriminator.apply(weights_init)
                print('Reloading dis net')

    def test(self, data=None, intrain=False, scale=True):
        # if not intrain:
        #     self.load_model(self.out_dir, "cur_w.pth")
        self.generator.eval()
        self.discriminator.eval()

        if data is None:
            dataloader = self.dataloader
        else:
            dataloader = preprocess_data(data, labels=None, param={}, is_train=False)
        
        rec_diff = self.get_diff(dataloader)
        # print(rec_diff)
        return rec_diff

    def save_cur_model(self, outdir, filename):
        if not os.path.exists(outdir):
            os.makedirs(outdir)

        state = {"generator": self.generator.state_dict(),
                 "discriminator": self.discriminator.state_dict(),
                 }
        torch.save(state, os.path.join(outdir, filename))

    def save_model_to(self, path):
        state = {"generator": self.generator.state_dict(),
                 "discriminator": self.discriminator.state_dict(),
                 }
        torch.save(state, path)

    def load_model(self, outdir, filename):
        state = torch.load(os.path.join(outdir, filename), map_location=self.device)
        self.generator.load_state_dict(state["generator"])
        self.discriminator.load_state_dict(state["discriminator"])

    def load_model_from(self, path):
        state = torch.load(path, map_location=self.device)
        self.generator.load_state_dict(state["generator"])
        self.discriminator.load_state_dict(state["discriminator"])

    def get_diff(self, dataloader):

        rec_diff = []
        ori_ts = []
        rec_ts = []
        rec_err = []
        with torch.no_grad():
            for i, data in enumerate(dataloader):
                data_X, y = data

                data_X = data_X.to(self.device)
                # data_cond = data_cond.to(self.device)

                rec_x, z = self.generator(data_X)

                mse_metric = torch.mean(torch.sum(torch.pow(data_X - rec_x, 2), dim=2),
                                        dim=1).detach().cpu().numpy()

                rec_diff.append(mse_metric)
                ori_ts.append(data_X.detach().cpu().numpy())
                rec_ts.append(rec_x.detach().cpu().numpy())
                rec_err.append(torch.sum(torch.pow(data_X - rec_x, 2), dim=2).detach().cpu().numpy())

        rec_diff = np.concatenate(rec_diff)
        ori_ts = np.concatenate(ori_ts)
        rec_ts = np.concatenate(rec_ts)
        rec_err = np.concatenate(rec_err)

        ori_ts = np.transpose(ori_ts, axes=(0, 2, 1))
        rec_ts = np.transpose(rec_ts, axes=(0, 2, 1))

        return rec_diff, ori_ts, rec_ts, rec_err


class BeatGAN_CNN(BeatGAN):
    def __init__(self, tensor, *args, **kwargs):
        super(BeatGAN_CNN, self).__init__(tensor)
        self.dataloader = preprocess_data(tensor, labels=None, param=kwargs, is_train=True)
        self.device = device
        self.lamda_value = param_default(kwargs, "lambda", param_default_dict)

        self.max_epoch = param_default(kwargs, "max_epoch", param_default_dict)

        self.generator = Generator(nc=param_default(kwargs, "input_size", param_default_dict),
                                   nz=param_default(kwargs, "rep_size", param_default_dict),
                                   seq_len=param_default(kwargs, "seq_len", param_default_dict), device=device).to(device)
        self.discriminator = Discriminator(nc=param_default(kwargs, "input_size", param_default_dict),
                                           seq_len=param_default(kwargs, "seq_len", param_default_dict), device=device).to(device)

        self.mse = nn.MSELoss()
        self.bce = nn.BCELoss()

        self.optimizerG = optim.Adam(self.generator.parameters(), lr=param_default(kwargs, "lr", param_default_dict))

        self.optimizerD = optim.Adam(self.discriminator.parameters(), lr=param_default(kwargs, "lr", param_default_dict))

        self.out_dir = param_default(kwargs, "out_path", param_default_dict)
        if not os.path.exists(self.out_dir):
            os.makedirs(self.out_dir)

        self.iteration = 0

        self.fix_input = None
        self.fix_data_cond = None

        if kwargs.__contains__("model_path"):
            self.load_model_from(kwargs["model_path"])

    def fit(self, data=None):
        return self.train(data)

    def predict(self, data=None):
        return self.test(data)

    def train(self, data=None):
        super().train(data)
        return self



import functools
import numbers

import numpy as np
import scipy.sparse as ssp
from scipy import signal
import sparse


def _ensure_array(x):
    if isinstance(x, DTensor):
        return x._data
    return np.asarray(x)


def _ensure_tensor(x):
    if isinstance(x, np.ndarray):
        if x.ndim == 0:
            return x.item()
        return DTensor(x)
    elif isinstance(x, sparse.SparseArray):
        return STensor.from_sparse_array(x)
    else:
        return x


def _wrap_ret(func):
    """Wrap return value of func to spartan tensor types.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        ret = func(*args, **kwargs)
        if type(ret) is tuple:
            return tuple(_ensure_tensor(x) for x in ret)
        else:
            return _ensure_tensor(ret)
    return wrapper


class DTensor(np.lib.mixins.NDArrayOperatorsMixin):
    """A dense multi-deimensional tensor on CPU (based on NumPy).

    Parameters
    ----------
    value : np.ndarray or array_like
        Data of dense tensor.
    dtype : data-type, optional
        Data type of the tensor, by default None(inferred automatically)

    Examples
    --------
    You can create a `DTensor` in many ways. For example, `numpy.ndarray`
    and `list of list`.

    >>> x = np.random.rand(3, 4)
    >>> x
    array([[0.1535191 , 0.4655889 , 0.07610213, 0.178842  ],
        [0.59293541, 0.74665444, 0.68131217, 0.38478126],
        [0.66008434, 0.54528835, 0.21170162, 0.99398061]])
    >>> A = st.DTensor(x)
    >>> A
    DTensor(
    array([[0.1535191 , 0.4655889 , 0.07610213, 0.178842  ],
        [0.59293541, 0.74665444, 0.68131217, 0.38478126],
        [0.66008434, 0.54528835, 0.21170162, 0.99398061]])
    )
    >>> A = st.DTensor.from_numpy(x)
    >>> A
    DTensor(
    array([[0.1535191 , 0.4655889 , 0.07610213, 0.178842  ],
        [0.59293541, 0.74665444, 0.68131217, 0.38478126],
        [0.66008434, 0.54528835, 0.21170162, 0.99398061]])
    )
    >>> A = st.DTensor([[1, 2, 3, 4], [2, 1, 4, 3], [1, 3, 2, 4]])
    >>> A
    DTensor(
    array([[1, 2, 3, 4],
        [2, 1, 4, 3],
        [1, 3, 2, 4]])
    )

    """

    def __init__(self, value, dtype=None):
        if isinstance(value, STensor):
            self._data = value._data.todense().astype(dtype)
        elif isinstance(value, DTensor):
            self._data = value._data.astype(dtype)
        else:
            self._data = np.asarray(value, dtype=dtype)

    _HANDLED_TYPES = (np.ndarray, numbers.Number)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        out = kwargs.get('out', ())
        for x in inputs + out:
            if not isinstance(x, self._HANDLED_TYPES + (DTensor,)):
                return NotImplemented

        # Defer to the implementation of the ufunc on unwrapped values.
        inputs = tuple(x._data if isinstance(x, DTensor) else x
                       for x in inputs)
        if out:
            kwargs['out'] = tuple(
                x._data if isinstance(x, DTensor) else x
                for x in out)
        result = getattr(ufunc, method)(*inputs, **kwargs)

        if type(result) is tuple:
            # multiple return values
            return tuple(type(self)(x) for x in result)
        elif method == 'at':
            # no return value
            return None
        else:
            # one return value
            return type(self)(result)

    def __repr__(self):
        return '%s(\n%r\n)' % (type(self).__name__, self._data)

    def __len__(self):
        return self._data.__len__()

    @_wrap_ret
    def __copy__(self):
        return self._data.__copy__()

    def __iter__(self):
        return iter(self._data)

    # Slice
    @_wrap_ret
    def __getitem__(self, index):
        return self._data.__getitem__(index)

    def __setitem__(self, index, value):
        self._data.__setitem__(index, value)

    def __delitem__(self, index):
        self._data.__delitem__(index)

    @_wrap_ret
    def __getattr__(self, name):
        if name.startswith('_'):
            return super().__getattr__(name)
        ret = getattr(self._data, name)
        if callable(ret):
            raise AttributeError(
                f"{type(self).__name__} doesn't have attribute {name}")
        return ret

    def __setattr__(self, name, value):
        if name.startswith('_'):
            super().__setattr__(name, value)
        else:
            setattr(self._data, name, value)

    def __delattr__(self, name):
        if name.startswith('_'):
            super().__delattr__(name)
        else:
            delattr(self._data, name)

    @_wrap_ret
    def all(self, axis=None, keepdims=False):
        """Wrapper of `numpy.all`

        Parameters
        ----------
        axis : None or int or tuple of ints , optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        DTensor or bool
            Result of `all` operation as a DTensor or a boolean value.
        """
        return np.all(self._data, axis=axis, keepdims=keepdims)

    @_wrap_ret
    def any(self, axis=None, keepdims=False):
        """Wrapper of `numpy.any`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        DTensor or bool
            Result of `any` operation as a DTensor or a boolean value.
        """
        return np.any(self._data, axis=axis, keepdims=keepdims)

    @_wrap_ret
    def min(self, axis=None, keepdims=False):
        """Wrapper of `numpy.ndarray.min`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        DTensor or scalar
            Result of `min` operation as a DTensor or a scalar value.
        """
        return self._data.min(axis=axis, keepdims=keepdims)

    @_wrap_ret
    def max(self, axis=None, keepdims=False):
        """Wrapper of `numpy.ndarray.max`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        DTensor or scalar
            Result of `max` operation as a DTensor or a scalar value.
        """
        return self._data.max(axis=axis, keepdims=keepdims)

    @_wrap_ret
    def sum(self, axis=None, keepdims=False):
        """Wrapper of `numpy.ndarray.sum`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        DTensor or scalar
            Result of `sum` operation as a DTensor or a scalar value.
        """
        return self._data.sum(axis=axis, keepdims=keepdims)

    @_wrap_ret
    def prod(self, axis=None, keepdims=False):
        """Wrapper of `numpy.ndarray.prod`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        DTensor or scalar
            Result of `prod` operation as a DTensor or a scalar value.
        """
        return self._data.prod(axis=axis, keepdims=keepdims)

    @_wrap_ret
    def mean(self, axis=None, keepdims=False):
        """Wrapper of `numpy.ndarray.mean`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        DTensor or scalar
            Result of `mean` operation as a DTensor or a scalar value.
        """
        return self._data.mean(axis=axis, keepdims=keepdims)

    @_wrap_ret
    def var(self, axis=None, keepdims=False):
        """Wrapper of `numpy.ndarray.var`, note that result is biased(divisor is `n`).

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are \
                left to size one, by default False

        Returns
        -------
        DTensor or scalar
            Result of `var` operation as a DTensor or a scalar value.
        """
        return self._data.var(axis=axis, keepdims=keepdims)

    @_wrap_ret
    def std(self, axis=None, keepdims=False):
        """Wrapper of `numpy.ndarray.std`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        DTensor or scalar
            Result of `std` operation as a DTensor or a scalar value.
        """
        return self._data.std(axis=axis, keepdims=keepdims)

    @_wrap_ret
    def dot(self, other):
        """Wrapper of `numpy.ndarray.dot`. Perform tensor dot operation with
        another tensor, including vector inner product, matrix multiplication
        and general tensor dot.

        Parameters
        ----------
        other : DTensor
            The second operand of dot operation.

        Returns
        -------
        DTensor or scalar
            Tensor dot result as a DTensor or a scalar value.
        """
        return self._data.dot(other._data)

    @_wrap_ret
    def reshape(self, shape):
        """Wrapper of `numpy.ndarray.reshape`. Gives a new dense tensor with
        new shape and current data.

        Parameters
        ----------
        shape : int or tuple of ints
            Shape of new tensor.

        Returns
        -------
        DTensor
            New dense tensor with given shape.
        """
        return self._data.reshape(shape)

    @_wrap_ret
    def nonzero(self):
        """Wrapper of `numpy.ndarray.nonzero`. Find the indices of non-zero elements.

        Returns
        -------
        tuple of DTensor
            Indices of non-zero elements.
        """
        return tuple([DTensor(d) for d in self._data.nonzero()])

    @_wrap_ret
    def astype(self, dtype):
        """Wrapper of `numpy.ndarray.astype`. Convert the array to a new type.

        Parameters
        ----------
        dtype : data-type
            Type of new tensor.

        Returns
        -------
        DTensor
            New dense tensor with given data type.
        """
        return self._data.astype(dtype)

    def to_numpy(self):
        """Return the tensor as `numpy.ndarray`.

        Returns
        -------
        numpy.ndarray
            Returned numpy array.
        """
        return self._data

    @classmethod
    def from_numpy(cls, x: np.ndarray):
        """Construct a `DTensor` from a `numpy.ndarray`.

        Parameters
        ----------
        x : numpy.ndarray
            Input array.

        Returns
        -------
        DTensor
            Dense tensor constructed from given array.

        Raises
        ------
        TypeError
            Raise if the parameter is not `numpy.ndarray`.
        """
        if not isinstance(x, np.ndarray):
            raise TypeError(
                f"Argument type should be `numpy.ndarray`, got {type(x)}")
        t = cls.__new__(cls)
        t._data = np.asarray(x)
        return t

    @_wrap_ret
    def resample(self, new_len: int, inplace: bool = False):
        """Resample DTensor to a new length.

        Parameters
        ----------
        new_len : int
            length of resampled tensor

        inplace : bool
            whether change on origin data

        Returns
        -------
        None or DTensor
            If not inplace, return resampled tensor.
        """
        _data = self._data
        if inplace:
            self._data = signal.resample(_data, new_len, axis=1)
        else:
            resampled_data = signal.resample(_data, new_len, axis=1)
            return resampled_data

    @_wrap_ret
    def concatenate(self, tensor: "DTensor" or np.ndarray, axis: int = 0, inplace: bool = False):
        """Concatenate DTensors together.

        Parameters
        ----------
        tensor : DTensor
            tensor to be concatenated

        axis : int
            dimension on which tenson will be concatenated

        inplace : bool
            whether change on origin data

        Returns
        ----------
        None or DTensor
            If not inplace, return concatenated tensor.
        """
        _type = type(tensor)
        if _type == self.__class__:
            _data = tensor._data
        elif _type == np.ndarray:
            _data = tensor
        if len(_data.shape) == 1:
            _data = [_data]
        if inplace:
            self._data = np.concatenate((self._data, _data), axis=axis)
        else:
            concatenated_data = np.concatenate((self._data, _data), axis=axis)
            return concatenated_data

    @_wrap_ret
    def cut(self, start: int, end: int, inplace: bool = True):
        if inplace:
            if len(self._data.shape) == 1:
                self._data = self._data[start: end]
            else:
                self._data = self._data[:, start: end]
        else:
            if len(self._data.shape) == 1:
                cut_data = self._data[start: end]
            else:
                cut_data = self._data[:, start: end]
            return cut_data


class STensor(np.lib.mixins.NDArrayOperatorsMixin):
    """A sparse multi-dimensional tensor on CPU (based on sparse).

    Parameters
    ----------
    data : numpy.ndarray or scipy.sparse.spmatrix or tuple of indices (ndim
    x nnz) and values (1 x nnz)
        Data of sparse tensor.
    shape : tuple, optional
        Shape of sparse tensor, by default None (inferred automatically)

    Examples
    --------
    You can create a `STensor` in many ways. For example, raw coordinates and
    values array, numpy.ndarray`, `scipy.spmatrix` and `sparse.COO`.

    >>> import numpy as np
    >>> import scipy.sparse as ssp
    >>> import sparse
    >>> import spartan as st
    >>> coords = np.array([[0, 1, 2], [2, 1, 0]])
    >>> vals = np.array([1, 2, 3])
    >>> A = st.STensor((coords, vals))
    >>> A
    STensor(<COO: shape=(3, 3), dtype=int32, nnz=3, fill_value=0>)
    >>> A.todense()
    DTensor(
    array([[0, 0, 1],
        [0, 2, 0],
        [3, 0, 0]])
    )
    >>> x = np.random.rand(3, 4)
    >>> x[x<0.8] = 0
    >>> A = st.STensor.from_scipy_sparse(ssp.csr_matrix(x))
    >>> A
    STensor(<COO: shape=(3, 4), dtype=float64, nnz=1, fill_value=0.0>)
    >>> A = st.STensor.from_numpy(x)
    >>> A
    STensor(<COO: shape=(3, 4), dtype=float64, nnz=1, fill_value=0.0>)
    >>> A = st.STensor.from_sparse_array(sparse.COO.from_numpy(x))
    >>> A
    STensor(<COO: shape=(3, 4), dtype=float64, nnz=1, fill_value=0.0>)
    """

    def __init__(self, data, shape=None):
        if type(data) is tuple:
            indices, values = data
            self._data = sparse.COO(indices, values, shape=shape)
            #import ipdb; ipdb.set_trace()
        elif isinstance(data, DTensor):
            self._data = sparse.as_coo(data._data, shape=shape)
        elif isinstance(data, STensor):
            self._data = data._data
        else:
            self._data = sparse.as_coo(data, shape=shape)

    _HANDLED_TYPES = (np.ndarray, numbers.Number)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        out = kwargs.get('out', ())
        for x in inputs + out:
            if not isinstance(x, self._HANDLED_TYPES + (STensor,)):
                return NotImplemented

        # Defer to the implementation of the ufunc on unwrapped values.
        inputs = tuple(x._data if isinstance(x, STensor) else x
                       for x in inputs)
        if out:
            kwargs['out'] = tuple(
                x._data if isinstance(x, STensor) else x
                for x in out)
        result = getattr(ufunc, method)(*inputs, **kwargs)

        if type(result) is tuple:
            # multiple return values
            return tuple(type(self)(x) for x in result)
        elif method == 'at':
            # no return value
            return None
        else:
            # one return value
            if isinstance(result, np.ndarray):
                return self.__class__.from_numpy(result)
            elif isinstance(result, ssp.spmatrix):
                return self.__class__.from_scipy_sparse(result)
            elif isinstance(result, sparse.COO):
                return self.__class__.from_sparse_array(result)
            else:
                return result

    def sum_to_scipy_sparse(self, modes: tuple = (0, 1)):
        m = len(self._data.shape)
        cmodes = tuple(set(range(m)) - set(modes))
        if len(cmodes) == 0:
            return self._data.to_scipy_sparse()
        return self._data.sum(axis=cmodes).to_scipy_sparse()

    def __repr__(self):
        return '%s(%r)' % (type(self).__name__, self._data)

    def __copy__(self):
        return self._data.__copy__()

    def __iter__(self):
        return iter(self._data)

    @_wrap_ret
    def __getattr__(self, name):
        if name.startswith('_'):
            return super().__getattr__(name)
        ret = getattr(self._data, name)
        if callable(ret):
            raise AttributeError(
                f"{type(self).__name__} doesn't have attribute {name}")
        return ret

    def __setattr__(self, name, value):
        if name.startswith('_'):
            super().__setattr__(name, value)
        else:
            setattr(self._data, name, value)

    def __delattr__(self, name):
        if name.startswith('_'):
            super().__delattr__(name)
        else:
            delattr(self._data, name)

    def __len__(self):
        return self._data.__len__()

    @_wrap_ret
    def __getitem__(self, index):
        return self._data.__getitem__(index)

    def __setitem__(self, index, value):
        self._data.__setitem__(index, value)

    def __delitem__(self, index):
        self._data.__delitem__(index)

    @_wrap_ret
    def all(self, axis=None, keepdims=False):
        """Wrapper of `numpy.all`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        STensor or bool
            Result of `all` operation as a STensor or a boolean value.
        """
        return np.all(self._data, axis=axis, keepdims=keepdims)

    @_wrap_ret
    def any(self, axis=None, keepdims=False):
        """Wrapper of `numpy.any`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        STensor or bool
            Result of `any` operation as a STensor or a boolean value.
        """
        return np.any(self._data, axis=axis, keepdims=keepdims)

    @_wrap_ret
    def min(self, axis=None, keepdims=False):
        """Wrapper of `sparse.COO.min`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        STensor or scalar
            Result of `min` operation as a STensor or a scalar value.
        """
        return self._data.min(axis=axis, keepdims=keepdims)

    @_wrap_ret
    def max(self, axis=None, keepdims=False):
        """Wrapper of `sparse.COO.max`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        STensor or scalar
            Result of `max` operation as a STensor or a scalar value.
        """
        return self._data.max(axis=axis, keepdims=keepdims)

    @_wrap_ret
    def sum(self, axis=None, keepdims=False):
        """Wrapper of `sparse.COO.sum`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        STensor or scalar
            Result of `sum` operation as a STensor or a scalar value.
        """
        return self._data.sum(axis=axis, keepdims=keepdims)

    @_wrap_ret
    def prod(self, axis=None, keepdims=False):
        """Wrapper of `sparse.COO.prod`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        STensor or scalar
            Result of `prod` operation as a STensor or a scalar value.
        """
        return self._data.prod(axis=axis, keepdims=keepdims)

    @_wrap_ret
    def mean(self, axis=None, keepdims=False):
        """Wrapper of `sparse.COO.mean`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        STensor or scalar
            Result of `mean` operation as a STensor or a scalar value.
        """
        return self._data.sum(axis=axis, keepdims=keepdims)

    @_wrap_ret
    def var(self, axis=None, keepdims=False):
        """Wrapper of `sparse.COO.var`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        STensor or scalar
            Result of `var` operation as a STensor or a scalar value.
        """
        return self._data.var(axis=axis, keepdims=keepdims)

    @_wrap_ret
    def std(self, axis=None, keepdims=False):
        """Wrapper of `sparse.COO.std`

        Parameters
        ----------
        axis : None or int or tuple of ints, optional
            Axis or axes to operate on, by default
            None
        keepdims : bool, optional
            If true, the axes along which the operation performed are left
            to size one, by default False

        Returns
        -------
        STensor or scalar
            Result of `std` operation as a STensor or a scalar value.
        """
        return self._data.std(axis=axis, keepdims=keepdims)

    @_wrap_ret
    def dot(self, other):
        """Wrapper of `sparse.COO.dot`. Perform tensor dot operation with
        another tensor, including vector inner product, matrix multiplication
        and general tensor dot.

        Parameters
        ----------
        other : DTensor or STensor
            The second operand of dot operation.

        Returns
        -------
        DTensor or STensor
            Tensor dot result as a DTensor or a STensor or a scalar value.
        """
        return self._data.dot(other._data)

    @_wrap_ret
    def todense(self):
        """Convert to a dense tensor.

        Returns
        -------
        DTensor
            Dense tensor coverted from the sparse tensor.
        """
        return self._data.todense()

    @_wrap_ret
    def reshape(self, shape):
        """Wrapper of `sparse.COO.reshape`. Gives a new dense tensor with
        new shape and current data.

        Parameters
        ----------
        shape : int or tuple of ints
            Shape of new tensor.

        Returns
        -------
        STensor
            New sparse tensor with given shape.
        """
        return self._data.reshape(shape)

    @_wrap_ret
    def nonzero(self):
        """Wrapper of `sparse.COO.nonzero`. Find the indices of non-zero elements.

        Returns
        -------
        tuple of DTensor
            Indices of non-zero elements.
        """
        return self._data.nonzero()

    @_wrap_ret
    def astype(self, dtype):
        """Wrapper of `sparse.COO.astype`. Convert the array to a new type.

        Parameters
        ----------
        dtype : data-type
            Type of new tensor.

        Returns
        -------
        STensor
            New sparse tensor with given data type.
        """
        return self._data.astype(dtype)

    def to_scipy(self, format='coo'):
        """Return the sparse tensor as a `scipy.sparse.spmatrix`.

        Parameters
        ----------
        format : str, optional, {'coo', 'csr', 'csc', 'lil', 'dok'}
            Format of scipy sparse matrix, by default 'coo'

        Returns
        -------
        scipy.sparse.spmatrix
            Returned scipy sparse matrix.
        """
        return self._data.to_scipy_sparse().asformat(format)

    def to_sparse_array(self):
        """Return the sparse tensor as a `sparse.COO`.

        Returns
        -------
        sparse.COO
            Returned sparse array.
        """
        return self._data

    @classmethod
    def from_numpy(cls, x: np.ndarray):
        """Construct a `STensor` from a `numpy.ndarray`.

        Parameters
        ----------
        x : numpy.ndarray
            Input array.

        Returns
        -------
        STensor
            Sparse tensor constructed from given array.

        Raises
        ------
        TypeError
            Raise if the parameter is not `numpy.ndarray`.
        """
        if not isinstance(x, np.ndarray):
            raise TypeError(
                f"Argument type should be `numpy.ndarray`, got {type(x)}")
        t = cls.__new__(cls)
        t._data = sparse.COO.from_numpy(x)
        return t

    @classmethod
    def from_scipy_sparse(cls, x: ssp.spmatrix):
        """Construct a `STensor` from a `scipy.sparse.spmatrix`.

        Parameters
        ----------
        x : scipy.sparse.spmatrix
            Input sparse matrix.

        Returns
        -------
        STensor
            Sparse tensor constructed from given sparse matrix.

        Raises
        ------
        TypeError
            Raise if the parameter is not `scipy.sparse.spmatrix`.
        """
        if not isinstance(x, ssp.spmatrix):
            raise TypeError(
                f"Argument type should be `scipy.sparse.spmatrix`, \
                got {type(x)}")
        t = cls.__new__(cls)
        t._data = sparse.COO.from_scipy_sparse(x)
        return t

    @classmethod
    def from_sparse_array(cls, x: sparse.COO):
        """Construct a sparse tensor from a `sparse.COO`.

        Parameters
        ----------
        x : sparse.COO
            Input COO sparse array.

        Returns
        -------
        STensor
            Sparse tensor constructed from given sparse array.

        Raises
        ------
        TypeError
            Raise if the parameter is not `sparse.COO`.
        """
        if not isinstance(x, sparse.COO):
            raise TypeError(
                f"Argument type should be `sparse.COO`, got {type(x)}")
        t = cls.__new__(cls)
        t._data = x
        return t



cdef extern from "math.h":
    double log2(double x)
cdef extern from "stdint.h":
    ctypedef int uint64_t


cpdef double log_comb(int n, int k):
    cdef int i = 0;
    cdef double ret = 0;
    while i < k:
        ret += log2(n-i);
        i += 1;
        ret -= log2(i);
    return ret


cpdef double LN(int n):
    """
    Encode length of an integer z by Rissanen's 1983 Universal code for integers
    """
    if n <= 0:
        return 0;
    cdef double ret = log2(2.865064);
    cdef double i = log2(n);
    while i > 0:
        ret += i;
        i = log2(i);
    return ret


cpdef double LnU(uint64_t n, uint64_t k):
    """
    Minimum encode length of n binary elements with k 1s and (n-k) 0s.
    """
    if n == 0 or k == 0 or k == n:
        return 0
    cpdef double x = -log2(<double> k / n);
    cpdef double y = -log2(<double> (n-k) / n);
    return k * x + (n-k) * y



import os,sys
from scipy.sparse import csc_matrix, coo_matrix, csr_matrix, lil_matrix


def checkfilegz(name):
    if os.path.isfile(name):
        return name
    elif os.path.isfile(name+'.gz'):
        return name+'.gz'
    else:
        return None

def myreadfile(fnm, mode):
    if '.gz' == fnm[-3:]:
        fnm = fnm[:-3]
    if os.path.isfile(fnm):
        f = open(fnm, mode)
    elif os.path.isfile(fnm+'.gz'):
        import gzip
        f = gzip.open(fnm+'.gz', mode)
    else:
        print('file: {} or its zip file does NOT exist'.format(fnm))
        sys.exit(1)
    return f

def saveSimpleDictData(simdict, outdata):
    with open(outdata, 'w') as fw:
        for k, v in simdict.iteritems():
            fw.write("{}:{}\n".format(k,v))
        fw.close()

def loadSimpleDictData(indata):
    simdict={}
    with myreadfile(indata, 'r') as fr:
        lines=fr.readlines()
        for line in lines:
            line = line.strip().split(':')
            simdict[int(line[0])]=float(line[1])
        fr.close()
    return simdict

def saveDictListData(dictls, outdata):
    with open(outdata, 'w') as fw:
        for k, l in dictls.iteritems():
            if type(l) != list:
                print("This is not a dict of value list.\n")
                break
            fw.write("{}:".format(k))
            for i in range(len(l)-1):
                fw.write("{} ".format(l[i]))
            fw.write("{}\n".format(l[-1]))
        fw.close()

def loadDictListData(indata, ktype=str, vtype=str):
    dictls={}
    with myreadfile(indata, 'r') as fr:
        lines = fr.readlines()
        for line in lines:
            line = line.strip().split(':')
            lst=[]
            for e in line[1].strip().split(' '):
                lst.append(vtype(e))
            dictls[ktype(line[0].strip())]=lst
        fr.close()
    return dictls

def saveSimpleListData(simls, outdata):
    with open(outdata, 'w') as fw:
        fw.write('\n'.join(map(str, simls)))
        fw.write('\n')
        fw.close()

def loadSimpleList(indata, dtype=None):
    sl=[]
    with myreadfile(indata, 'r') as fi:
        for e in fi:
            e = e.strip()
            if e == '':
                continue
            if dtype is not None:
                e = dtype(e)
            sl.append(e)
        fi.close()
    return sl

def save2DarrayData(sarray,outdata):
    with open(outdata, 'w') as fw:
        for l in sarray:
            if len(l) > 0:
                for i in range(len(l)-1):
                    fw.write("{} ".format(l[i]))
                fw.write("{}\n".format(l[-1]))
        fw.close()

def load2DarrayData(indata):
    a2d=[]
    with myreadfile(indata, 'r') as fi:
        for l in fi:
            arr = l.strip().split(' ')
            a2d.append(arr)
        fi.close()
    return a2d

def scsmatrix2edgelist(scsm, weighted=False):
    el = []
    rows, cols = scsm.nonzero()
    l = len(rows)
    for i in range(l):
        if weighted == True:
            el.append((rows[i], cols[i], scsm[rows[i], cols[i]]))
        else:
            el.append((rows[i], cols[i]))
    return el

def writeedgelist(el, ofile, weighted=False):
    out = open(ofile, "wb")
    for e in el:
        if weighted==True:
            out.write("%d %d %f\n" % (e[0], e[1], e[2]))
        else:
            out.write("%d %d\n" % (e[0], e[1]))
    out.close()

def readedge2squarecscm(ifile, weighted=False, delimiter=' '):
    xs=[]
    ys=[]
    data=[]
    with myreadfile(ifile, 'rb') as fin:
        for line in fin:
            coords=line.strip().split(delimiter)
            xs.append(int(coords[0]))
            ys.append(int(coords[1]))
            if weighted:
                data.append(float(coords[2]))
            else:
                data.append(1.0)
        fin.close()
    M = max(max(xs), max(ys)) +1
    cscm = csc_matrix((data, (xs, ys)), shape=(M,M))
    return cscm

def cscmatrix2groupedges(scsm):
    ges = []
    rows, cols = scsm.nonzero()
    l = len(rows)
    for i in range(l):
        ges.append((rows[i], cols[i]))
    return ges

def writegroupedges(ges, ofile):
    out = open(ofile, "wb")
    for ge in ges:
        out.write("%d %d\n" % ge)
    out.close()

def readedge2coom(ifile, weighted=False, delimiter=' ', idstartzero=True):
    xs=[]
    ys=[]
    data=[]
    if idstartzero is True:
        offset = 0
    else:
        offset = -1
    with myreadfile(ifile, 'rb') as fin:
        for line in fin:
            coords=line.strip().split(delimiter)
            xs.append( int(coords[0]) + offset )
            ys.append( int(coords[1]) + offset )
            if weighted:
                data.append(float(coords[2]))
            else:
                data.append(1)
        fin.close()
    M = max(xs) + 1
    N = max(ys) + 1
    coom = coo_matrix((data, (xs, ys)), shape=(M,N))
    return coom

def loadedge2sm(ifile, mtype=csc_matrix, dtype=int, delimiter=' ',
                idstartzero=True, issquared=False):
    '''
    load edge list into sparse matrix
    matrix dimensions are decided by max row id and max col id
    support csr, coo, csc matrix
    '''
    xs=[]
    ys=[]
    data=[]
    if idstartzero is True:
        offset = 0
    else:
        offset = -1
    with myreadfile(ifile, 'rb') as fin:
        for line in fin:
            coords=line.strip().split(delimiter)
            xs.append( int(coords[0]) + offset )
            ys.append( int(coords[1]) + offset )
            if len(coords) == 3:
                data.append(dtype(coords[2]))
            else:
                data.append(1)
        fin.close()
    m = max(xs) + 1
    n = max(ys) + 1
    if issquared is False:
        M, N = m, n
    else:
        M = max(m,n)
        N = max(m,n)
    sm = mtype((data, (xs, ys)), shape=(M,N))
    return sm

def savesm2edgelist(sm, ofile, idstartzero=True, delimiter=' ', kformat=None):
    '''
    Save sparse matrix into edgelist
    kformat = "{0:.1f}"
    '''
    offset = 0 if idstartzero else 1
    lsm = sm.tolil()
    with open(ofile, 'wb') as fout:
        i = 0
        for row, d in zip(lsm.rows, lsm.data):
            for j,w in zip(row, d):
                if kformat is not None:
                    w = str.format(kformat, w)
                ostr = delimiter.join([str(i+offset), str(j+offset), str(w)])
                #fout.write('{} {} {}\n'.format(i+offset, j+offset, w))
                fout.write(ostr+'\n')
            i+=1
        fout.close()
    return


if __name__== "__main__":
    path= './testdata/'
    ifile = path+"yelp.edgelist"
    sm = loadedge2sm(ifile, csr_matrix, idstartzero=True)



#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

#  Project: eaglemine
#      ploter.py
#		  visualization tools
#      Created by @wenchieh  on <11/25/2017>


__author__ = 'wenchieh'

# sys
import warnings
warnings.filterwarnings("ignore")
import collections as clct

# third-party lib
import numpy as np
import matplotlib
matplotlib.use("Agg")
import matplotlib.cm as cm
import matplotlib.pylab as plt
from matplotlib.colors import LogNorm


# Plot item degree distribution (frequency / P(x>=d)) vs. degree
def plot_degree_dist(degree, value, xlabel, ylabel, color_marker='b.',
                     title=None, outfn=None):
    fig = plt.figure()
    plt.loglog(degree, value, color_marker, markersize=2, markeredgecolor=None)
    plt.xlabel(xlabel, linespacing=12, fontsize=18)
    plt.ylabel(ylabel, linespacing=12, fontsize=18)
    plt.tight_layout()
    if title is not None: plt.title(title, fontsize=18, y=1.05)
    if outfn is not None:
        fig.savefig(outfn)
        plt.close()
    return fig


def plot_log2pdf(feature, xlabel, color_marker='b.', ylabel='Frequency', outfn=None):
    ft2cnt = dict()
    for f in feature:
        ft2cnt[f] = ft2cnt.get(f, 0) + 1

    fig = plt.figure()
    plt.loglog(ft2cnt.keys(), ft2cnt.values(), color_marker)
    plt.xlabel(xlabel, linespacing=12, fontsize=18)
    plt.ylabel(ylabel, linespacing=12, fontsize=18)
    plt.tight_layout()
    if outfn is not None:
        plt.savefig(outfn)
        plt.close()
    # plt.show()
    return fig


def plot_clusters(data, center_pts, data_labels, outliers=list([]),
                  core_samples=None, grid=False, ticks=True, outfn=None):
    fig = plt.figure(figsize=(6.5, 6), dpi=96)
    lab2cnt = clct.Counter(data_labels)
    cmap = cm.get_cmap('Spectral')
    colors = cmap(np.linspace(0, 1, len(lab2cnt)))
    if core_samples is not None:
        core_samples_mask = np.zeros_like(data_labels, dtype=bool)
        core_samples_mask[core_samples] = True
    else:
        core_samples_mask = np.ones_like(data_labels, dtype=bool)

    N_clusters = 0
    keys, values = np.array(list(lab2cnt.keys())), np.array(list(lab2cnt.values()))
    srt_lab = keys[np.argsort(values)][::-1]
    for k, col in zip(srt_lab, colors):
        if k == -1:
            # Black used for noise.
            # col = 'gray'
            continue

        N_clusters += 1
        class_member_mask = (data_labels == k)
        xy = data[class_member_mask & core_samples_mask]
        if len(xy) > 0:
            plt.plot(xy[:, 1], xy[:, 0], 's', color=col, markersize=6) #markerfacecolor=col, markeredgecolor=col

        mn_xy = np.mean(xy, 0)
        plt.text(mn_xy[1], mn_xy[0], str(k),
                 {'color': 'k', 'fontsize': 18, 'ha': 'center', 'va': 'center',
                  'bbox': dict(boxstyle="circle", fc="w", ec="k", pad=0.2, alpha=0.3)} )
        xy = data[class_member_mask & (np.invert(core_samples_mask))]
        if len(xy) > 0:
            plt.plot(xy[:, 1], xy[:, 0], 's', color=col, markersize=6) #markerfacecolor=col, markeredgecolor=col
        if len(center_pts) > 0:
            plt.plot(center_pts[k, 0], center_pts[k, 1], 's',
                     markerfacecolor=col, markeredgecolor='k', markersize=10)

    ubnd, rbnd = np.max(np.vstack((data, outliers)), 0) + 1
    if len(outliers) > 0:
        plt.plot(outliers[:, 1], outliers[:, 0], 'bo', markersize=1)
        uo, ro = np.max(outliers, 0)
        rbnd, ubnd = np.max([rbnd, ro]), np.max([ubnd, uo])

    ax = plt.gca()
    ax.tick_params(axis='both', which='both', bottom='off', left='off',
                   labelbottom='off', labelleft='off', )
    # ax.set_axis_bgcolor('white')   # deprecated method in Matplotlib v2.0
    ax.set_facecolor('white')

    spine_linewidth = 6 #8#
    for spine in ax.spines.values():
        spine.set_linewidth(spine_linewidth)
    plt.grid(grid)
    plt.xlim((0, rbnd + 0.2))
    plt.ylim((0, ubnd + 0.2))
    plt.tight_layout()
    if outfn is not None:
        plt.savefig(outfn)
        plt.close()
    # plt.show()
    return fig


def plot_hexbin_heatmap(xs, ys, xlabel, ylabel, xscale='log',
                        yscale='log', gridsize=100, colorscale=True, outfn=None):
    '''
    xscale: [ ‘linear’ | ‘log’ ]
        Use a linear or log10 scale on the horizontal axis.
    yscale: [ ‘linear’ | ‘log’ ]
        Use a linear or log10 scale on the vertical axis.
    gridsize: [ 100 | integer ]
        The number of hexagons in the x-direction, default is 100. The
        corresponding number of hexagons in the y-direction is chosen such that
        the hexagons are approximately regular. Alternatively, gridsize can be
        a tuple with two elements specifying the number of hexagons in the
        x-direction and the y-direction.
    '''
    fig = plt.figure()

    if colorscale:
        plt.hexbin(xs, ys, bins='log', gridsize=gridsize, xscale=xscale,
                   yscale=yscale, mincnt=1, cmap=cm.get_cmap('jet'))
        cb = plt.colorbar()
        cb.set_label(r'$log_{10}(N)$', fontsize=16)
    else:
        plt.hexbin(xs, ys, gridsize=gridsize, xscale=xscale, yscale=yscale,
                   mincnt=1, cmap=cm.get_cmap('jet'))
        cb = plt.colorbar()
        cb.set_label('counts')
    plt.xlabel(xlabel, linespacing=12, fontsize=18)
    plt.ylabel(ylabel, linespacing=12, fontsize=18)
    plt.tight_layout()

    if outfn is not None:
        fig.savefig(outfn)
        plt.close()
    # plt.show()
    return fig


def plot_heatmap(x_vec, y_vec, heatmap, base=10, xlabel=None, ylabel=None, outfn=None):
    n, m = heatmap.shape
    fig = plt.figure(figsize=(8, 6.5), dpi=96)
    plt.pcolormesh(heatmap, cmap='jet', norm=LogNorm(), rasterized=True)
    cb = plt.colorbar()
    for lb in cb.ax.yaxis.get_ticklabels():
        lb.set_family('Times New roman')
        lb.set_size(20)

    ax = fig.gca()
    xticks = ax.get_xticks()
    if xticks[-1] > m:  xticks = xticks[:-1]
    xstep = xticks[1] - xticks[0]
    nw_xtick = []
    for xt in xticks:
        if (xt < m) and (xt % (2*xstep) == 0):
            pws = int(np.log(x_vec[int(xt)]) / np.log(base))
            if pws != 0:
                nw_xtick.append('%dE%d' % (base, pws))
            else:
                nw_xtick.append('1')
        else:
            nw_xtick.append('')

    nw_ytick = []
    for yt in ax.get_yticks():
        if yt < n:
            yval = y_vec[int(yt)]
            if yval < 1e4:
                nw_ytick.append(r'%d' % yval)
            else:
                pws = int(np.log10(yval))
                fv = yval * 1.0 / 10**pws
                nw_ytick.append('%.1fE%d'%(fv, pws))

    if nw_xtick[-1] == '':
        nw_xtick[-1] = '%.2f'%x_vec[-1]
        # nw_xtick[-1] = '%.2f'%np.power(base, x_vec[-1])
    if nw_ytick[-1] == '':
        nw_ytick[-1] = '%d' % int(y_vec[-1])
        # nw_ytick = '%d' % int(np.power(base, y_vec[-1]))

    ax.set_xticklabels(nw_xtick, fontsize=27, family='Times New roman')
    ax.set_yticklabels(nw_ytick, fontsize=27, family='Times New roman')

    if xlabel is not None:
        plt.xlabel(xlabel, linespacing=12, fontsize=32, family='Times New roman')
    if ylabel is not None:
        plt.ylabel(ylabel, linespacing=12, fontsize=32, family='Times New roman')

    # fig.set_size_inches(8, 7.3)
    fig.tight_layout()
    if outfn is not None:
        fig.savefig(outfn)
    return fig


# plot the ROC/AUC curves given #N True-Positive-Rate & False-Positive-Rate Series
#     colors, markers for each serie
def plot_roc_curve(NTPRS, NFPRS, legends_str, colors, markers, outfn=None):
    # colors=['r', 'g', 'm', 'b', 'c', 'k'],
    # markers=[None, None, None, None, None, None, '>', 'v', '+']
    N_med = len(NTPRS)
    fig = plt.figure()
    for k in range(N_med):
        plt.plot(NFPRS[k], NTPRS[k], color=colors[k], linewidth=2,
                 linestyle='-', marker=markers[k], label=legends_str[k])
    # plt.plot([0, 1], [0, 1], color='navy', linewidth=2, linestyle='--')
    plt.plot([0, 1], [0, 1], color='k', linewidth=2, linestyle='--')
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.0])

    plt.xlabel('False positive rate', linespacing=12, fontsize=32, family='Times New roman', fontweight='bold')
    plt.ylabel('True positive rate', linespacing=12, fontsize=32, family='Times New roman', fontweight='bold')
    plt.legend(loc='lower right', ncol=1, fancybox=False, shadow=False, fontsize=22)
    plt.tight_layout()
    if outfn is not None:
        plt.savefig(outfn)
        plt.close()
    # plt.show()
    return fig


def plot_heatmap_graphlab_pgrk(x_vec, y_vec, heatmap, base=10, xlabel=None, ylabel=None, outfn=None):
    n, m = heatmap.shape
    fig = plt.figure(figsize=(8, 6.5), dpi=96)
    plt.pcolormesh(heatmap, cmap='jet', norm=LogNorm(), rasterized=True)

    cb = plt.colorbar()
    for lb in cb.ax.yaxis.get_ticklabels():
        lb.set_family('Times New roman')
        lb.set_size(20)

    ax = fig.gca()
    xticks = ax.get_xticks()
    if xticks[-1] > m:  xticks = xticks[:-1]
    ### for graph-lab  pagerank featrue
    nw_xtick = []
    for xt in xticks: #ax.get_xticks():
        if ((xt < m) and (xt % 40 == 0)):
            xval = x_vec[int(xt)]
            if xval < 1:
                nw_xtick.append(r'%.2f' % xval)
            elif xval < 100:
                nw_xtick.append(r'%.1f' % xval)
            else:
                pws = int(np.log10(xval))
                fv = xval * 1.0 / 10**pws
                nw_xtick.append(r'%.1fE%d'%(fv, pws))
        else:
            nw_xtick.append('')

    nw_ytick = []
    for yt in ax.get_yticks():
        if yt < n:
            yval = y_vec[int(yt)]
            if yval < 1e3:
                nw_ytick.append(r'%d' % yval)
            else:
                pws = int(np.log10(yval))
                fv = yval * 1.0 / 10**pws
                nw_ytick.append('%.1fE%d'%(fv, pws))

    if nw_xtick[-1] == '':
        nw_xtick[-1] = '%.2f'%x_vec[-1]
        # nw_xtick[-1] = '%.2f'%np.power(base, x_vec[-1])
    if nw_ytick[-1] == '':
        nw_ytick[-1] = '%d' % int(y_vec[-1])
        # nw_ytick = '%d' % int(np.power(base, y_vec[-1]))

    ax.set_xticklabels(nw_xtick, fontsize=23, family='Times New roman')
    ax.set_yticklabels(nw_ytick, fontsize=23, family='Times New roman')

    if xlabel is not None:
        plt.xlabel(xlabel, linespacing=12, fontsize=32, family='Times New roman')
    if ylabel is not None:
        plt.ylabel(ylabel, linespacing=12, fontsize=32, family='Times New roman')

    # fig.set_size_inches(8, 7.3)
    fig.tight_layout()
    if outfn is not None:
        fig.savefig(outfn)
        plt.close()
    return fig


def plot_heatmap_2discretes(x_vec, y_vec, heatmap, base=10, xlabel=None, ylabel=None, outfn=None):
    n, m = heatmap.shape
    fig = plt.figure(figsize=(8, 6.5), dpi=96)
    plt.pcolormesh(heatmap, cmap='jet', norm=LogNorm(), rasterized=True)
    cb = plt.colorbar()

    for l in cb.ax.yaxis.get_ticklabels():
        l.set_family('Times New roman')
        l.set_size(20)

    ax = fig.gca()
    xticks = ax.get_xticks()
    if xticks[-1] > m:  xticks = xticks[:-1]

    ### for triangle_count / degree  feature
    nw_xtick = []
    for xt in xticks:
        if ((xt < m) and (xt % 20 == 0)):
            xval = x_vec[int(xt)]
            if xval < 1e3:
                nw_xtick.append(r'%d' % xval)
            else:
                pws = int(np.log10(xval))
                fv = xval * 1.0 / 10**pws
                nw_xtick.append(r'%.1fE%d'%(fv, pws))
        else:
            nw_xtick.append('')

    nw_ytick = []
    for yt in ax.get_yticks():
        if yt < n:
            yval = y_vec[int(yt)]
            if yval < 1e3:
                nw_ytick.append(r'%d' % yval)
            else:
                pws = int(np.log10(yval))
                fv = yval * 1.0 / 10**pws
                nw_ytick.append('%.1fE%d'%(fv, pws))

    if nw_xtick[-1] == '':
        nw_xtick[-1] = '%.2f'%x_vec[-1]
        # nw_xtick[-1] = '%.2f'%np.power(base, x_vec[-1])
    if nw_ytick[-1] == '':
        nw_ytick[-1] = '%d' % int(y_vec[-1])
        # nw_ytick = '%d' % int(np.power(base, y_vec[-1]))

    ax.set_xticklabels(nw_xtick, fontsize=27, family='Times New roman')
    ax.set_yticklabels(nw_ytick, fontsize=27, family='Times New roman')

    if xlabel is not None:
        plt.xlabel(xlabel, linespacing=12, fontsize=32, family='Times New roman')
    if ylabel is not None:
        plt.ylabel(ylabel, linespacing=12, fontsize=32, family='Times New roman')

    # fig.set_size_inches(8, 7.3)
    fig.tight_layout()
    if outfn is not None:
        fig.savefig(outfn)
        plt.close()
    return fig



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   __init__.py
@Desc    :   Import part and configuration part.
'''

# here put the import lib


from spartan.util import *
from spartan.tensor import TensorData



#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

# #  Example tool for extracting graph nodes feature
#  Author: wenchieh
#
#  Project: eaglemine
#      graph.py
#      Version:  1.0
#      Date: December 17 2017
#      Main Contact: Wenchieh Feng (wenchiehfeng.us@gmail.com)
#
#      Copyright:
#        This software is free of charge under research purposes.
#        For commercial purposes, please contact the author.
#
#      Created by @wenchieh  on <12/17/2017>
#

__author__ = 'wenchieh'

# sys
from abc import ABCMeta, abstractmethod

# third-party lib
import numpy as np
import networkx as nx
from scipy.sparse import lil_matrix
from scipy.sparse.linalg import svds


class BasicGraph(object):
    __metaclass__ = ABCMeta

    # default setting to control the precision/error of alternative iterative updated SVD
    __ITERS_MAX = 20
    __ITER_PRECISION = 1e-20

    def __init__(self, weighted=False):
        self.weighted = weighted
        self.n_src, self.n_dest = 0, 0
        self.n_edges = 0
        self.edgelist = None
        self.src_hub, self.dest_auth = None, None

    def set_edgelist(self, edgelist_data):
        self.edgelist = edgelist_data
        self.n_edges = len(self.edgelist)
    
    def _get_sparse_matrix(self, matrix_type=lil_matrix):
        m, n = self.get_number_of_node()
        sps_mat = matrix_type((m, n), dtype=float)
        for i, j in self.edgelist:
            if self.weighted:
                if [i, j] not in sps_mat:
                    sps_mat[i, j] = 1.0
                else:
                    sps_mat[i, j] += 1.0
            else:
                sps_mat[i, j] = 1.0
        return sps_mat

    def get_node_degree(self):
        '''
        get the out-degree (in-degree) for source (destination) nodes.
        :return:
        '''
        m, n = self.get_number_of_node()
        self.src_outd, self.dest_ind = np.zeros(m, int), np.zeros(n, int)
        unique_edges = set()

        for i, j in self.edgelist:
            if self.weighted:
                self.src_outd[i] += 1
                self.dest_ind[j] += 1
            else:
                if (i, j) not in unique_edges:
                    self.src_outd[i] += 1
                    self.dest_ind[j] += 1
                    unique_edges.add((i, j))
        return self.src_outd, self.dest_ind

    def _set_hits(self, hub, auth):
        '''
        set the hits score based on given hubness [hub] and authority [auth] containing some post-processing
        :param hub: hubness array (len(hub) == # src_node)
        :param auth: authority aray (len(auth) == # dest_node)
        :return:
        '''
        if abs(np.max(hub)) < abs(np.min(hub)):   # len(np.where(hub <= 0)[0]) >= 0.5 * len(hub):
            hub *= -1
        hub[hub < 0] = 0
        if abs(np.max(auth)) < abs(np.min(auth)): # len(np.where(auth <= 0)[0]) >= 0.5 * len(auth):
            auth *= -1
        auth[auth < 0] = 0

        self.src_hub = hub
        self.dest_auth = auth

    def get_hits_score_alternate(self, iters=15, error=None):
        '''
        calculate HITS score (hub score u1 and authority score v1) for the graph iteratively,
        with [T] iterations or [error] iteration error (precision), one of them should be given at least.
        :param iters: the maximum number of iterations (default=15)
        :param error: the minimum iterate error.
        :return:
        '''
        hub = np.zeros(self.n_src)
        auth = np.ones(self.n_dest) * 1.0 / np.sqrt(self.n_dest)  # with mode = 1

        iters_max = self.__ITERS_MAX
        if (iters is not None) and (iters > 0):
            iters_max = min([self.__ITER_PRECISION, iters])

        error_min = self.__ITER_PRECISION
        if (error is not None) and (error > 0):
            error_min = max([self.__ITER_PRECISION, error])

        # each iteration of HITS algorithm
        k = 0
        while True:
            hub_pre = np.array(hub)
            for e in self.edgelist:
                hub[e[0]] += auth[e[1]]
            hub /= np.sqrt(np.sum(hub ** 2))
            for e in self.edgelist:
                auth[e[1]] += hub[e[0]]
            auth /= np.sqrt(np.sum(auth ** 2))

            iter_error = np.mean(np.abs(hub_pre - hub))
            if (k > iters_max) or (iter_error <= error_min):
                print("iterations: {}, error: {}".format(k, iter_error))
                break
            k += 1

        self._set_hits(hub, auth)
        return self.src_hub, self.dest_auth

    def get_hits_score(self):
        '''
        get HITS score feature for nodes of graph.
        using the SVD algorithm (maybe slowly)
        :return:
        '''
        sps_mat = self._get_sparse_matrix()
        hub, s, auth = svds(sps_mat, k=1, which='LM')
        hub = np.squeeze(np.array(hub))
        auth = np.squeeze(np.array(auth))
        self._set_hits(hub, auth)
        return self.src_hub, self.dest_auth

    @abstractmethod
    def get_number_of_node(self):
        '''
        get the number of the node in graph
        :return:
        '''
        pass

    def get_neighbor_associativity(self):
        '''
        get the neighbors associativity feature (average degree of the 1st-order neighbors) for each nodes of graph.
        :return:
        '''
        self.src_assort = np.zeros(self.n_src, int)
        self.dest_assort = np.zeros(self.n_dest, int)

        for i, j in self.edgelist:
            self.src_assort[i] += self.dest_ind[j]
            self.dest_assort[j] += self.src_outd[i]

        for k in range(self.n_src):
            if self.src_outd[k] > 0:
                self.src_assort[k] = int(self.src_assort[k] / self.src_outd[k])

        for k in range(self.n_dest):
            if self.dest_ind[k] > 0:
                self.dest_assort[k] = int(self.dest_assort[k] / self.dest_ind[k])
        return self.src_assort, self.dest_assort

    @abstractmethod
    def get_corenum(self):
        '''
        get the core number of each node of graph.
        '''
        pass

    def save_features(self, out_fn, src_score, dest_score, sep, header=None):
        m, n = self.n_src, self.n_dest
        with open(out_fn, 'w') as fp:
            fp.writelines('# %d%s%d\n' % (m, sep, n))
            if header is not None:
                fp.writelines('# out/in degree, ' + header + '\n')
            for k in range(m):
                str_line = str(self.src_outd[k]) + sep + str(src_score[k])
                fp.writelines(str_line + '\n')

            for k in range(n):
                str_line = str(self.dest_ind[k]) + sep + str(dest_score[k])
                fp.writelines(str_line + '\n')
            fp.close()

    def save_edgelist(self, out_fn, sep=','):
        with open(out_fn, 'w') as fp:
            for e in self.edgelist:
                str_line = str(e[0]) + sep + str(e[1])
                fp.writelines(str_line + '\n')
            fp.close()

    def save_deg2hits(self, out_fn, sep=','):
        '''
        save HITS feature to file.
        format: ([out-degree][sep][hub], [[in-degree][sep][authority]])
        :param out_fn: output file name
        :param sep: separator of the file (default=',')
        :return:
        '''
        self.save_features(out_fn, self.src_hub, self.dest_auth, sep, "hubness/authority")

    def save_deg2corehum(self, out_fn, sep=','):
        '''
        save the core-number feature to file.
        format: [out(in)-degree][sep][score_number]
        :param out_fn: output file name
        :param sep: separator of the file (default=',')
        :return:
        '''
        self.save_features(out_fn, self.src_cores, self.dest_cores, sep, "src/sink corenum")

    def save_deg2ngbassort(self, out_fn, sep=','):
        '''
        save neighbor associativity feature to file.
        format: [out(in)-degree][sep][neighbor-associativity]
        :param out_fn: output file name
        :param sep: separator of the file (default=',')
        :return:
        '''
        header = "avg in/out degree of 1-st neighbor for src/sink node resp."
        self.save_features(out_fn, self.src_assort, self.dest_assort, sep, )


class BipartiteGraph(BasicGraph):
    def __init__(self, weighted=False):
        super(BasicGraph, self).__init__()
        self.weighted = weighted

    def get_number_of_node(self):
        self.n_src = np.max(self.edgelist[:, 0]) + 1
        self.n_dest = np.max(self.edgelist[:, 1]) + 1
        return self.n_src, self.n_dest

    def get_corenum(self):
        '''
        get the core-number feature of each node.
        construct bipartite with networkx to calculate core-number of each node.
        :return:
        '''
        g = nx.Graph()
        g.add_nodes_from(range(self.n_src), bipartite=0)
        g.add_nodes_from(range(self.n_src, self.n_dest + self.n_src), bipartite=1)
        edges = np.array(self.edgelist)
        edges[:, 1] += self.n_src
        g.add_edges_from(edges)
        cores = nx.core_number(g).values()
        self.src_cores = cores[:self.n_src]
        self.dest_cores = cores[self.n_src:]
        return self.src_cores, self.dest_cores


class UnipartiteGraph(BasicGraph):
    def __init__(self, weighted=False):
        super(BasicGraph, self).__init__()
        self.weighted = weighted

    def set_direct(self, directed=True):
        self.directed = directed

    def set_edgelist(self, edgelist_data):
        self.n_edges = len(edgelist_data)
        if self.directed == False:
            rev_edges = edgelist_data[:, ::-1]
            edgelist_data = np.vstack((edgelist_data, rev_edges))
        self.edgelist = edgelist_data

    def get_node_degree(self):
        sp_mat = self._get_sparse_matrix()
        g = nx.DiGraph()
        xs, ys = sp_mat.nonzero()
        for k in range(len(xs)):
            g.add_edge(xs[k], ys[k], weight=sp_mat[xs[k], ys[k]])
        self.degree = np.asarray(g.degree().values())
        return self.degree

    def get_number_of_node(self):
        max_id = np.max(self.edgelist)
        self.n_src = self.n_dest = max_id + 1
        return self.n_src, self.n_dest

    def get_corenum(self):
        '''
        get the core-number feature of each node.
        construct unipart-graph with networkx to calculate core-number of each node.
        :return:
        '''
        g = nx.Graph()
        g.add_edges_from(self.edgelist)
        self.src_cores = self.dest_cores = nx.core_number(g).values()
        return self.src_cores, self.dest_cores

    def get_triangle(self):
        '''
        get triangle number feature for each node of graph with can be hold only for unipart-graph.
        construct unipart-graph with networkx for calculating the triangles
        :return:
        '''
        g = nx.Graph()
        g.add_edges_from(self.edgelist)
        self.nds_triangles = nx.triangles(g).values()
        return self.nds_triangles

    def get_associate_deg(self):
        '''
        get associate feature of each node of directed unipart-graph (out-degree vs. in-degree.)
        :return:
        '''
        if self.directed == True:
            self.ass_degs = np.zeros((self.n_src, 2), int)
            self.ass_degs[:, 0] = self.src_outd
            self.ass_degs[:, 1] = self.dest_ind
            return self.ass_degs
        else:
            ValueError("No associate features: Only the node of directed "
                       "homogeneous graph have both in-degree and out-degree!")
            return None

    def get_pagerank(self):
        sp_mat = self._get_sparse_matrix()
        g = nx.DiGraph()
        xs, ys = sp_mat.nonzero()
        for k in range(len(xs)):
            g.add_edge(xs[k], ys[k], weight=sp_mat[xs[k], ys[k]])
        self.pagerank = np.asarray(nx.pagerank(g).values())
        return self.pagerank

    def save_deg2pgrk(self, out_fn, sep=','):
        with open(out_fn, 'r') as fp:
            fp.writelines("# #node:{}, #edges:{}, #feature:{}\n".format(self.n_src, self.n_edges, 2))
            fp.writelines("# degree, pagerank\n")
            for k in range(self.n_src):
                line = "{}{}{}".format(self.degree[k], sep, self.pagerank[k])
                fp.writelines(line + '\n')
            fp.close()

    def save_associate_deg(self, out_fn, sep=','):
        with open(out_fn, 'w') as fp:
            fp.writelines("# #node:{}, #edges:{}, #feature:{}\n".format(self.n_src, self.n_edges, 2))
            fp.writelines("# out-degree, in-degree\n")
            for k in range(len(self.ass_degs)):
                line = "{}{}{}".format(self.ass_degs[k, 0], sep ,str(self.ass_degs[k, 1]))
                fp.writelines(line + '\n')
            fp.close()

    def save_deg2triangles(self, out_fn, sep=','):
        '''
        save node triangles-number feature to file.
        format: ([out-degree][sep][triangles], [[in-degree][sep][triangles]])
        :param out_fn: output file name
        :param sep: separator of the file (default=',')
        :return:
        '''
        with open(out_fn, 'w') as fp:
            fp.writelines("# #node:{}, #edges:{}, #feature:{}\n".format(self.n_src, self.n_edges, 2))
            fp.writelines("# degree, #triangles\n")
            for k in range(self.n_src):
                line = "{}{}{}".format(self.degree[k], sep, self.nds_triangles[k])
                fp.writelines(line + '\n')
            fp.close()



#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

# #  The Goodness-of-fit (GoF) for truncated normal distribution
#  Author: wenchieh
#
#  Project: eaglemine
#      truncate_norm_gof.py
#      Version:  1.0
#      Date: December 21 2017
#      Main Contact: Wenchieh Feng (wenchiehfeng.us@gmail.com)
#
#      Copyright:
#        This software is free of charge under research purposes.
#        For commercial purposes, please contact the author.
#
#      Created by @wenchieh  on <12/21/2017>
#

__author__ = 'wenchieh'


# sys
import errno
import multiprocessing
from collections import Counter
from multiprocessing import Process
from multiprocessing.queues import Queue

# third-party lib
import numpy as np
from scipy.stats import norm, uniform
import scipy.optimize as opt


def _ad_alters_check(alternative):
    valid_alters = ["twosided", "less", "greater"]
    if alternative not in valid_alters:
        raise ValueError("Error: Invalid argument for alternative hypothesis, "
                   "which must be on of {}".format(valid_alters))

def _ad_stat_(n, zH, zs, alternative='twosided'):
    '''
    Supremum Class Anderson-Darling test
           Supremum class version of the Anderson-Darling test providing a comparison of a
         fitted distribution with the empirical distribution.
    :param n: the size of samples  #pts
    :param zH: the probability of point at truncate-clip point H, i.e. zH = \head F_{theta} (H)
    :param zs: the probability of each samples i.e. zs_j = \head F_{theta} (x_j)
               |zs| = n
    :param alternative: indicates the alternative hypothesis and must be one of
                    "twosided" (default), "less", or "greater".
                    Initial letter must be specified only.
    :return: supremum AD statistic
    '''
    assert(len(zs) == n)
    _ad_alters_check(alternative)

    AD = 0.0
    if alternative == 'greater':
        part = np.array([(zH + 1.0*(j+1)*(1-zH)/n - zs[j])/(np.sqrt((zs[j]-zH)*(1-zs[j]))) for j in range(n)])
        AD = np.sqrt(n) * np.max(part)
    elif alternative == 'less':
        part = np.array([(zs[j] - zH - 1.0*(j-1)*(1-zH)/n)/(np.sqrt((zs[j]-zH)*(1-zs[j]))) for j in range(n)])
        AD = np.sqrt(n) * np.max(part)
    else:
        part_p = np.array([(zH + 1.0*j*(1-zH)/n - zs[j])/(np.sqrt((zs[j]-zH)*(1-zs[j]))) for j in range(n)])
        part_m = np.array([(zs[j] - zH - 1.0*(j-1)*(1-zH)/n)/(np.sqrt((zs[j]-zH)*(1-zs[j]))) for j in range(n)])
        AD = np.sqrt(n) * np.max([part_p, part_m])

    return AD

def _aduptail_stat_(n, zH, zs, alternative='twosided'):
    '''
    Supremum Class Upper Tail Anderson-Darling test
           Supremum class version of the Upper Tail Anderson-Darling test providing a comparison
         of a fitted distribution with the empirical distribution.
    :param n: the size of samples  #pts
    :param zH: the probability of point at truncate-clip point H, i.e. zH = \head F_{theta} (H)
    :param zs: the probability of each samples i.e. zs_j = \head F_{theta} (x_j)
               |zs| = n
    :param alternative: indicates the alternative hypothesis and must be one of
                    "twosided" (default), "less", or "greater".
                    Initial letter must be specified only.
    :return: supremum AD statistic
    '''
    assert(len(zs) == n)
    _ad_alters_check(alternative)

    ADUP = 0.0
    if alternative == 'greater':
        part = np.array([(zH + 1.0*(1-zH)*(j+1)/n - zs[j])/(1-zs[j]) for j in range(n)])
        ADUP = np.sqrt(n) * np.max(part)
    elif alternative == 'less':
        part = np.array([(zs[j] - zH - 1.0*j*(1-zH)/n)/(1-zs[j]) for j in range(n)])
        ADUP = np.sqrt(n) * np.max(part)
    else:
        part_p = np.array([(zH + 1.0*(1-zH)*(j+1)/n - zs[j])/(1-zs[j]) for j in range(n)])
        part_m = np.array([(zs[j] - zH - 1.0*j*(1-zH)/n)/(1-zs[j]) for j in range(n)])
        ADUP = np.sqrt(n) * np.max([part_p, part_m])

    return ADUP

def _ad2_stat_(n, zH, zs):
    '''
    the Quadratic Class Anderson-Darling Test Statistic for left-truncated data
           Quadratic class Anderson-Darling test providing a comparison of a fitted
         distribution with the empirical distribution.
    :param n:  the size of samples  #pts
    :param zH: the probability of point at truncate-clip point H, i.e. zH = \head F_{theta} (H)
    :param zs: the probability of each samples i.e. zs_j = \head F_{theta} (x_j)
               |zs| = n
    :return: AD2 statistic
    '''
    assert(len(zs) == n)
    part = np.sum([(1+2*(n-(j+1)))*np.log(1-zs[j]) - (1-2*(j + 1))*np.log(zs[j] - zH) for j in range(n)])
    AD2 = -n + 2*n*np.log(1-zH) - part * 1.0 / n
    return AD2

def _ad2uptail_stat_(n, zH, zs):
    '''
    the Quadratic Class Upper Tail Anderson-Darling test
           Quadratic Class Upper Tail Anderson-Darling test providing a comparison of a fitted
        distribution with the empirical distribution.
    :param n:  the size of samples  #pts
    :param zH: the probability of point at truncate-clip point H, i.e. zH = \head F_{theta} (H)
    :param zs: the probability of each samples i.e. zs_j = \head F_{theta} (x_j)
               |zs| = n
    :return: AD2UPTAIL statistic
    '''
    assert(len(zs) == n)
    part = np.sum([(1 + 2*(n-(j+1)))/(1 - zs[j]) for j in range(n)])
    AD2UP = -2*n*np.log(1-zH) + 2*np.sum(np.log(1 - zs)) + (1.0 - zH) / n * part
    return AD2UP

def retry_on_eintr(function, *args, **kw):
    while True:
        try:
            return function(*args, **kw)
        except IOError as e:
            if e.errno == errno.EINTR:
                continue
            else:
                raise


class RetryQueue(Queue):
    def __init__(self):
        super(RetryQueue, self).__init__(ctx=multiprocessing.get_context())

    def get(self, block=True, timeout=None):
        return retry_on_eintr(Queue.get, self, block, timeout)


class TruncatedNormalGoF(object):
    '''
    Reference:
        Chernobay, A., Rachev, S., Fabozzi, F. (2005), Composites goodness-of-fit tests
            for left-truncated loss samples, Tech. rep., University of Calivornia Santa Barbara
    '''

    def __init__(self, n_jobs=1):
        self._n_jobs_ = n_jobs

    def _optimize_mle_(self, init_pram, xs, weights, H, epsilon=1e-10):
        n = len(xs)
        mean, std =  init_pram[0], init_pram[1]

        dens = 1 - norm.cdf(H, loc=mean, scale=std) + epsilon
        numerator = norm.pdf(xs, loc=mean, scale=std)
        xs_pdf = numerator / dens
        wtxs_pdf = weights * xs_pdf
        val = np.asarray(wtxs_pdf[xs > H])
        if val.all() > 0:
            res = -1.0 * np.sum(np.log(val))
        else:
            res = -1.0 * np.NaN

        if np.isinf(res):
            res = - np.log(np.finfo('float64').resolution) * n

        return res

    def _sample_fit_(self, xs, init_mean, init_std, H, weights=None,
                     tol=1e-4, maxiter=4000, verbose=False):
        n = len(xs)

        if weights is None:
            weights = np.array([1] * n)
        init_parm = np.array([init_mean, init_std])
        res = opt.minimize(self._optimize_mle_, init_parm, args=(xs, weights, H),
                           method = 'BFGS', tol = tol,
                           options={'maxiter': maxiter, 'disp': verbose})

        mean, std = res.x[0], res.x[1]
        return mean, std


    def _mc_simulate_(self, n, mean, std, zH, H, STATSITIC_func, sims,
                      multi=False, queue=0, jobno=0):
        np.random.seed(jobno)
        sz = list()
        # 1. generate large number of samples (sim) from the fitted truncated distribution of size n
        for i in range(sims):
            uniszs = uniform.rvs(size=n)
            sz_k = norm.ppf(uniszs*(1 - zH) + zH, loc=mean, scale=std)
            sz.append(sz_k)
        # 2. Fit truncated distribution and estimate conditional parameters $\head \theta$
        mc_TS = list()
        for i in range(sims):
            mc_mean, mc_std = self._sample_fit_(sz[i], mean, std, H)

        # 3. Estimate the GOF statistic value Di for each sample
            mc_zH = norm.cdf(H, loc=mc_mean, scale=mc_std)

            uni_szi = sorted(np.unique(sz[i]))
            count = Counter(sz[i])
            mc_zs = list()
            for mczj in uni_szi:
                mczj_cdf = norm.cdf(mczj, loc=mc_mean, scale=mc_std)
                mc_zs.extend([mczj_cdf] * count[mczj])
            mc_zs = np.asarray(mc_zs)
            # mc_zs = np.array([norm.cdf(mczj, loc=mc_mean, scale=mc_std) for mczj in sorted(sz[i])])
            mc_TS.append(STATSITIC_func(len(sz[i]),mc_zH, mc_zs))

        if multi:
            queue.put(mc_TS)
        else:
            return mc_TS

    def _multimc_simus_(self, n, mean, std, zH, H, STATSITIC_func, sims):
        res = list()
        queues = [RetryQueue() for i in range(self._n_jobs_)]
        jobs_sims = int(np.ceil(sims * 1.0 / self._n_jobs_))
        args = [(n, mean, std, zH, H, STATSITIC_func, jobs_sims, True, queues[i], i + 1)
                for i in range(self._n_jobs_)]
        jobs = [Process(target=self._mc_simulate_, args=(a)) for a in args]
        for j in jobs: j.start()
        for q in queues: res.append(q.get())
        for j in jobs:  j.join()

        return np.hstack(res)


    def _montecarlo_test_(self, xs, mean, std, H, STATSITIC_func, sims, tol):
        '''
        Monte-Carlo simulation based GoF test
               Performs Monte-Carlo based Goodness-of-Fit tests.
               the mctest is called by the GoF tests defined in this package. For internal use only.
        :return:
        '''
        if min(xs) < H:
            raise ValueError("'min(xs)' must be greater or equal to 'H'")

        p_value = 0.0
        n = len(xs)
        zH = norm.cdf(H, loc=mean, scale=std)
        uni_xs = sorted(np.unique(xs))
        count = Counter(xs)
        zs = list()
        for zj in uni_xs:
            zjcdf = norm.cdf(zj, loc=mean, scale=std)
            zs.extend([zjcdf] * count[zj])
        zs = np.asarray(zs)
        # zs = np.array([norm.cdf(zj, loc=mean, scale=std) for zj in sorted(xs)])
        TS0 = STATSITIC_func(n, zH, zs)

        if not np.isfinite(TS0):  #(np.abs(TS0) == np.inf) or (TS0 is np.NaN):
            # raise ValueError("test statistic value can't be calculated")
            print("Warning: test statistic value can't be calculated")
            return p_value, TS0, 0

        # # 1. generate large number of samples (sims) from the fitted truncated distribution of size n
        # np.random.seed(1024)
        # sz = list()
        # for i in range(sims):
        #     uniszs = uniform.rvs(size=n)
        #     sz_k = norm.ppf(uniszs*(1 - zH) + zH, loc=mean, scale=std)
        #     sz.append(sz_k)
        # # 2. Fit truncated distribution and estimate conditional parameters $\head \theta$
        # mc_TS = list()
        # for i in range(sims):
        #     mc_mean, mc_std = self._sample_fit_(sz[i], mean, std, H)
        #
        # # 3. Estimate the GOF statistic value Di for each sample
        #     mc_zH = norm.cdf(H, loc=mc_mean, scale=mc_std)
        #     mc_zs = np.array([norm.cdf(mczj, loc=mc_mean, scale=mc_std) for mczj in sorted(sz[i])])
        #     mc_TS.append(STATSITIC_func(len(sz[i]),mc_zH, mc_zs))

        mc_TS = list()
        # 1. generate large number of samples (sims) from the fitted truncated distribution of size n
        # 2. Fit truncated distribution and estimate conditional parameters $\head \theta$
        # 3. Estimate the GOF statistic value Di for each sample
        if self._n_jobs_ > 1:
            mc_TS = self._multimc_simus_(n, mean, std, zH, H, STATSITIC_func, sims)
        else:
            mc_TS = self._mc_simulate_(n, mean, std, zH, H, STATSITIC_func, sims)

        # 4. Calculate p-value as the proportion of times the sample statistic values
        #     exceed the observed value D of the original samples
        k, z0, z1 = 0, 0, 2
        suc_nsims = 0
        for i in range(len(mc_TS)):
            if mc_TS[i] > TS0:
                k += 1
            z0 = z1
            z1 = k * 1.0 / (i + 1)
            suc_nsims += 1
            if (np.abs(z1 - z0) < tol) and (np.abs(z1 - 0.5) < 0.5):
                break
        p_value = z1 + 1e-20

        # 5. Reject H0 if the p-value is smaller than \alpha
        return p_value, TS0, suc_nsims


    def ad_test(self, xs, mean, std, H=np.NaN, alternative='twosided', sim=100, tol=1e-4):
        _ad_alters_check(alternative)
        if H is np.NAN:
            H = -np.inf

        def ad_stat(n, zH, zs):
            return _ad_stat_(n, zH, zs, alternative)

        p_value, TS, nsims = self._montecarlo_test_(xs, mean, std, H, ad_stat, sim, tol)

        ans = {'method': 'Supremum Class Anderson-Darling Test',
               'statitic_name': 'AD', 'statistic': TS, 'alternative': alternative,
               'p_value': p_value, 'nsims': nsims, 'threshold': H}

        return ans

    def adup_test(self, xs, mean, std, H=np.NaN, alternative='twosided',
                sim=100, tol=1e-4):
        _ad_alters_check(alternative)
        if H is np.NAN:
            H = -np.inf

        def aduptail_stat(n, zH, zs):
            return _aduptail_stat_(n, zH, zs, alternative)

        p_value, TS, nsims = self._montecarlo_test_(xs, mean, std, H, aduptail_stat, sim, tol)

        stat_name = { 'twosided': 'ADup', 'less': 'ADup-', 'greater': 'ADup+'  }
        ans = {'method': 'Anderson-Darling Upper Tail Test',
               'statistic_name': stat_name[alternative], 'statistic': TS, 'alternative': alternative,
               'p_value': p_value, 'nsims': nsims, 'threshold': H}
        return ans

    def ad2_test(self, xs, mean, std, H=np.NaN, sim=100, tol=1e-4):
        if H is np.NAN:
            H = -np.inf
        p_value, TS, nsims = self._montecarlo_test_(xs, mean, std, H, _ad2_stat_, sim, tol)
        ans = {'method':  'Quadratic Class Anderson-Darling Test',
               'statistic_name': 'AD2', 'statistic': TS,
               'p_value': p_value, 'nsims': nsims, 'threshold': H}
        return ans

    def ad2up_test(self, xs, mean, std, H=np.NaN, sim=100, tol=1e-4):
        if H is np.NAN:
            H = -np.inf
        p_value, TS, nsims = self._montecarlo_test_(xs, mean, std, H, _ad2uptail_stat_, sim, tol)
        ans = {'method':  'Quadratic Class Anderson-Darling Test',
               'statistic_name': 'AD2up', 'statistic': TS,
               'p_value': p_value, 'nsims': nsims, 'threshold': H}
        return ans



import torch
import os
import numpy as np
import torch.nn as nn
import torch.optim as optim
from tqdm import tqdm
from .metric import evaluate
from .preprocess import preprocess_data
from .._model import MLmodel

from .import param_default

os.environ["CUDA_VISIBLE_DEVICES"] = "0"
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
class Encoder(nn.Module):
    def __init__(self, layers, input_size, hidden_size, cond_size, rep_size, device, cell='gru'):
        super(Encoder, self).__init__()
        self.device = device
        self.layers = layers
        self.cell = cell

        self.input_size = input_size
        self.hidden_size = hidden_size
        self.cond_size = cond_size
        self.rep_size = rep_size

        if cell == "rnn":
            self.rnn1 = nn.RNNCell(self.input_size+self.cond_size, self.hidden_size)
            self.rnn2 = nn.ModuleList([nn.RNNCell(self.hidden_size, self.hidden_size) for i in range(self.layers - 1)])
        elif cell == "lstm":
            self.rnn1 = nn.LSTMCell(self.input_size+self.cond_size, self.hidden_size)
            self.rnn2 = nn.ModuleList([nn.LSTMCell(self.hidden_size, self.hidden_size) for i in range(self.layers - 1)])
        elif cell == "gru":
            self.rnn1 = nn.GRUCell(self.input_size + self.cond_size, self.hidden_size)
            self.rnn2 = nn.ModuleList([nn.GRUCell(self.hidden_size, self.hidden_size) for i in range(self.layers - 1)])
        else:
            raise Exception("no this cell:{}".format(cell))

        self.linear = nn.Linear(self.hidden_size, self.rep_size)

    def forward(self, input, cond):
        '''
        :param input:  batch,seq,ts dim
        :param cond: batch,seq,cond dim
        :return:
        '''

        h_t = torch.zeros(input.size(0), self.hidden_size, device=self.device)
        c_t = torch.zeros(input.size(0), self.hidden_size, device=self.device)
        if self.cell == "lstm":
            hc_all = [(h_t, c_t)]
        else:
            hc_all = [h_t]
        for i in range(self.layers-1):
            h_t2 = torch.zeros(input.size(0), self.hidden_size, device=self.device)
            c_t2 = torch.zeros(input.size(0), self.hidden_size, device=self.device)
            if self.cell == "lstm":
                hc_all.append((h_t2, c_t2))
            else:
                hc_all.append(h_t2)
        for i in range(input.size(1)):
            input_t = input[:, i, :]
            if self.cond_size > 0:
                x = torch.cat([input_t, cond[:, i, :]], dim=1)
            else:
                x = input_t
            hc_all[0] = self.rnn1(x, hc_all[0])
            for i in range(1, self.layers):
                if self.cell == "lstm":
                    hc_all[i] = self.rnn2[i-1](hc_all[i-1][0], hc_all[i])
                else:
                    hc_all[i] = self.rnn2[i - 1](hc_all[i - 1], hc_all[i])
        if self.cell == "lstm":
            z = self.linear(hc_all[-1][0])
        else:
            z = self.linear(hc_all[-1])
        return z


class Decoder(nn.Module):
    def __init__(self, layers, input_size, hidden_size, cond_size, rep_size, device, cell="gru"):
        '''
        :param input_size:  time series dims
        :param hidden_size:
        :param cond_size:  condition dims
        '''
        super(Decoder, self).__init__()
        self.device = device
        self.layers = layers
        self.cell = cell

        self.input_size = input_size
        self.hidden_size = hidden_size
        self.cond_size = cond_size
        self.rep_size = rep_size

        if cell == "rnn":
            self.rnn1 = nn.RNNCell(self.input_size+self.cond_size, self.hidden_size)
            self.rnn2 = nn.ModuleList([nn.RNNCell(self.hidden_size, self.hidden_size) for i in range(self.layers - 1)])
        elif cell == "lstm":

            self.rnn1 = nn.LSTMCell(self.input_size+self.cond_size, self.hidden_size)
            self.rnn2 = nn.ModuleList([nn.LSTMCell(self.hidden_size, self.hidden_size) for i in range(self.layers - 1)])
        elif cell == "gru":
            self.rnn1 = nn.GRUCell(self.input_size + self.cond_size, self.hidden_size)
            self.rnn2 = nn.ModuleList([nn.GRUCell(self.hidden_size, self.hidden_size) for i in range(self.layers - 1)])
        else:
            raise Exception("no this cell:{}".format(cell))

        self.linear_z = nn.Linear(self.rep_size, self.input_size)
        self.linear_out = nn.Linear(self.hidden_size, self.input_size)
        self.tanh = nn.Tanh()

    def forward(self, seq_len, init_hidden, cond, target=None):
        '''

        :param init_hidden:
        :param cond:
        :param target:
        :return:
        '''
        cur_input = self.linear_z(init_hidden)

        h_t = torch.zeros(init_hidden.size(0), self.hidden_size, device=self.device)
        c_t = torch.zeros(init_hidden.size(0), self.hidden_size, device=self.device)
        if self.cell == "lstm":
            hc_all = [(h_t, c_t)]
        else:
            hc_all = [h_t]

        for i in range(self.layers-1):
            h_t2 = torch.zeros(init_hidden.size(0), self.hidden_size, device=self.device)
            c_t2 = torch.zeros(init_hidden.size(0), self.hidden_size, device=self.device)
            if self.cell == "lstm":
                hc_all.append((h_t2, c_t2))
            else:
                hc_all.append(h_t2)

        outputs = []

        for i in range(seq_len):

            if self.cond_size > 0:
                cond_t = cond[:, i, :]
                x = torch.cat([cur_input, cond_t], dim=1)
            else:
                x = cur_input

            hc_all[0] = self.rnn1(x, hc_all[0])
            for i in range(1, self.layers):
                if self.cell == "lstm":
                    hc_all[i] = self.rnn2[i - 1](hc_all[i - 1][0], hc_all[i])
                else:
                    hc_all[i] = self.rnn2[i - 1](hc_all[i - 1], hc_all[i])

            if self.cell == "lstm":
                cur_input = self.tanh(self.linear_out(hc_all[-1][0]))
            else:
                cur_input = self.tanh(self.linear_out(hc_all[-1]))

            outputs += [cur_input]
        outputs = torch.stack(outputs, 1)
        return outputs


class Discriminator(nn.Module):
    def __init__(self, layers, input_size, hidden_size, cond_size, device, cell="lstm"):
        super(Discriminator, self).__init__()
        self.device = device
        self.layers = layers
        self.cell = cell

        self.input_size = input_size
        self.hidden_size = hidden_size
        self.cond_size = cond_size

        if cell == "rnn":
            self.rnn1 = nn.RNNCell(self.input_size+self.cond_size, self.hidden_size)
            self.rnn2 = nn.ModuleList([nn.RNNCell(self.hidden_size, self.hidden_size) for i in range(self.layers - 1)])
        elif cell == "lstm":

            self.rnn1 = nn.LSTMCell(self.input_size+self.cond_size, self.hidden_size)
            self.rnn2 = nn.ModuleList([nn.LSTMCell(self.hidden_size, self.hidden_size) for i in range(self.layers - 1)])
        elif cell == "gru":
            self.rnn1 = nn.GRUCell(self.input_size + self.cond_size, self.hidden_size)
            self.rnn2 = nn.ModuleList([nn.GRUCell(self.hidden_size, self.hidden_size) for i in range(self.layers - 1)])
        else:
            raise Exception("no this cell:{}".format(cell))

        self.linear = nn.Linear(self.hidden_size, 1)
        self.sigmoid = nn.Sigmoid()

    def forward(self, input, cond):
        '''

        :param input:  batch,seq,ts dim
        :param cond: batch,seq,cond dim
        :return:
        '''

        h_t = torch.zeros(input.size(0), self.hidden_size, device=self.device)
        c_t = torch.zeros(input.size(0), self.hidden_size, device=self.device)

        if self.cell == "lstm":
            hc_all = [(h_t, c_t)]
        else:
            hc_all = [h_t]
        for i in range(self.layers-1):
            h_t2 = torch.zeros(input.size(0), self.hidden_size, device=self.device)
            c_t2 = torch.zeros(input.size(0), self.hidden_size, device=self.device)
            if self.cell == "lstm":
                hc_all.append((h_t2, c_t2))
            else:
                hc_all.append(h_t2)

        for i in range(input.size(1)):
            input_t = input[:, i, :]
            if self.cond_size > 0:
                x = torch.cat([input_t, cond[:, i, :]], dim=1)
            else:
                x = input_t
            hc_all[0] = self.rnn1(x, hc_all[0])
            for i in range(1, self.layers):
                if self.cell == "lstm":
                    hc_all[i] = self.rnn2[i - 1](hc_all[i - 1][0], hc_all[i])
                else:
                    hc_all[i] = self.rnn2[i - 1](hc_all[i - 1], hc_all[i])
        if self.cell == "lstm":
            logit = self.linear(hc_all[-1][0])
            output = self.sigmoid(logit)
            feature = hc_all[-1][0]
        else:
            logit = self.linear(hc_all[-1])
            output = self.sigmoid(logit)
            feature = hc_all[-1]

        return output, logit, feature


class BeatGAN(MLmodel):
    def __init__(self, data, *args, **param):
        super(BeatGAN, self).__init__(data, *args, **param)
        self.dataloader = preprocess_data(data, False, param)
        self.device = device
        self.lamda_value = param_default(param, "lambda", 1)
        self.seq_len = param_default(param, "seq_len", 64)
        self.max_epoch = param_default(param, "max_epoch", 5)
        self.lr = param_default(param, "lr", 0.01)

        self.encoder = Encoder(param_default(param, "layers", 1), \
            param_default(param, "input_size", 1), param_default(param, "hidden_size", 100), \
                0, param_default(param, "rep_size", 20), device, cell=param_default(param, "net_type", "gru")).to(device)
        self.decoder = Decoder(param_default(param, "layers", 1), \
            param_default(param, "input_size", 1), param_default(param, "hidden_size", 100), \
                0, param_default(param, "rep_size", 20), device, cell=param_default(param, "net_type", "gru")).to(device)
        self.discriminator = Discriminator(param_default(param, "layers", 1), \
            param_default(param, "input_size", 1), param_default(param, "hidden_size", 100), \
                0, device, cell=param_default(param, "net_type", "gru")).to(device)

        self.mse = nn.MSELoss()
        self.bce = nn.BCELoss()
        self.l1loss = nn.L1Loss()

        params = list(self.encoder.parameters()) + list(self.decoder.parameters())
        self.optimizerG = optim.Adam(params, lr=self.lr)
        self.optimizerD = optim.Adam(self.discriminator.parameters(), lr=self.lr)

        self.out_dir = os.path.join("output", self.model_name)
        if not os.path.exists(self.out_dir):
            os.makedirs(self.out_dir)

        self.best_loss = float("inf")
        self.best_loss_epoch = 0
        self.early_stop = 10
        self.iteration = 0

        self.fix_input = None
        self.fix_data_cond = None

    def train(self):
        self.encoder.train()
        self.decoder.train()
        self.discriminator.train()

        best_auc = 0
        best_auc_epo = 0
        best_f1 = 0
        best_f1_epo = 0

        for epoch in tqdm(range(self.max_epoch)):
            self.train_epoch()

            self.save_cur_model(self.out_dir, "cur_w.pth")

            # print(
            #     "[{}]test auc:{:.4f} th:{:.4f} f1:{:.4f} \n".format(epoch, test_auc, test_th, test_f1))
            # if epoch-self.best_loss_epoch>self.early_stop:
            #     print("early stop!!!")
            #     break
            if epoch > 0 and epoch % 10 == 0:
                for param_group in self.optimizerG.param_groups:
                    param_group['lr'] = param_group['lr'] * 0.75
                for param_group in self.optimizerD.param_groups:
                    param_group['lr'] = param_group['lr'] * 0.75

            # print("best auc:{} in [{}]    best f1:{}  in [{}]".format(best_auc, best_auc_epo, best_f1, best_f1_epo))

    def train_epoch(self):
        self.encoder.train()
        self.decoder.train()
        self.discriminator.train()

        for i, data in enumerate(self.dataloader):
            self.iteration += 1

            data_X = data
            data_cond = None

            data_X = data_X.to(self.device)

            if self.fix_input is None:
                self.fix_input = data_X
                self.fix_data_cond = data_cond

            # train discriminator
            # self.discriminator.zero_grad()
            self.optimizerD.zero_grad()

            # train with real

            d_real_prob, d_real_logit, d_real_feat = self.discriminator(data_X, data_cond)

            # train with fake
            z = self.encoder(data_X, data_cond)
            rec_x = self.decoder(self.seq_len, z, data_cond)
            d_fake_prob, d_fake_logit, d_fake_feat = self.discriminator(rec_x, data_cond)

            loss_d_real = self.bce(d_real_prob, torch.full((d_real_prob.size(0),), 1, device=self.device))

            loss_d_fake = self.bce(d_fake_prob, torch.full((d_fake_prob.size(0),), 0, device=self.device))

            loss_d = loss_d_real+loss_d_fake
            loss_d.backward()
            self.optimizerD.step()

            # train generator
            # self.encoder.zero_grad()
            # self.decoder.zero_grad()
            self.optimizerG.zero_grad()

            z = self.encoder(data_X, data_cond)
            rec_x = self.decoder(self.seq_len, z, data_cond)
            d_real_prob, d_real_logit, d_real_feat = self.discriminator(data_X, data_cond)
            d_fake_prob, d_fake_logit, d_fake_feat = self.discriminator(rec_x, data_cond)

            loss_g_rec = self.mse(data_X, rec_x)
            loss_g_adv = self.mse(d_real_feat, d_fake_feat)
            # loss_g_adv=self.bce(d_fake_prob,torch.full((d_fake_prob.size(0),), 1,device=self.device))

            loss_g = loss_g_rec+loss_g_adv*self.lamda_value
            loss_g.backward()
            self.optimizerG.step()

            if (i % 100) == 0:
                print("{}:loss_g(rec/adv):{}/{},loss_d(real/fake):{}/{}".format(i, loss_g_rec, loss_g_adv*self.lamda_value, loss_d_real, loss_d_fake))
                # z = self.encoder(self.fix_input, self.fix_data_cond)
                # rec_x = self.decoder(self.seq_len,z, self.fix_data_cond)
                # show_num=8
                # img_tensor=multi_ts2image(rec_x.detach().cpu().numpy()[:show_num],self.fix_input.cpu().numpy()[:show_num])
                # for img_idx in range(show_num):
                #     self.writer.add_image("image/rec"+str(img_idx),img_tensor[img_idx], global_step=self.iteration, dataformats='HWC')

    def test(self, intrain=False):
        # if not intrain:
        #     self.load_model(self.out_dir,"cur_w.pth")
        self.encoder.eval()
        self.decoder.eval()
        self.discriminator.eval()

        rec_diff = self.get_diff(self.dataloader, save_pic=False)
        print(rec_diff)
        return rec_diff

    def save_cur_model(self, outdir, filename):
        if not os.path.exists(outdir):
            os.makedirs(outdir)

        state = {"encoder": self.encoder.state_dict(), "decoder": self.decoder.state_dict(),
                 "discriminator": self.discriminator.state_dict(),
                 }
        torch.save(state, os.path.join(outdir, filename))

    def save_model_to(self, path):
        state = {"encoder": self.encoder.state_dict(), "decoder": self.decoder.state_dict(),
                 "discriminator": self.discriminator.state_dict(),
                 }
        torch.save(state, path)

    def load_model(self, outdir, filename):
        state = torch.load(os.path.join(outdir, filename))
        self.encoder.load_state_dict(state["encoder"])
        self.decoder.load_state_dict(state["decoder"])
        self.discriminator.load_state_dict(state["discriminator"])

    def load_model_from(self, path):
        state = torch.load(path)
        self.encoder.load_state_dict(state["encoder"])
        self.decoder.load_state_dict(state["decoder"])
        self.discriminator.load_state_dict(state["discriminator"])

    def get_diff(self, dataloader, save_pic=False):

        rec_diff = []
        data_cond = None
        with torch.no_grad():
            for i, data in enumerate(dataloader):
                data_X = data

                data_X = data_X.to(self.device)

                z = self.encoder(data_X, data_cond)
                rec_x = self.decoder(self.seq_len, z, data_cond)

                mse_metric = torch.mean(torch.sum(torch.pow(data_X - rec_x, 2), dim=2),
                                        dim=1).detach().cpu().numpy()
                rec_diff.append(mse_metric)

        rec_diff = np.concatenate(rec_diff)

        return rec_diff

class BeatGAN_RNN(BeatGAN):
    def __init__(self, *args, **kwargs):
        super(BeatGAN_RNN, self).__init__(*args, **kwargs)

    def fit(self):
        self.train()
    
    def predict(self):
        self.test()

    def train(self):
        super().train()
        return self



import re
import pandas as pd
import numpy as np
import networkx as nx
import os
import pdb
import sparse
import scipy.sparse as ssp
import timeit

from ...tensor import TensorData
from .. import CubeFlow

# Remember to add spartan to you PATH
# import sys
# sys.path.append("/<dir to spartan2>/spartan2")
# sys.path.append("./spartan2-FCC")
# import spartan as st

def divide_connected_conponents(res, amt_tensor, cmt_tensor):
    subgraph_connected_list = []
    subgraph_timebin_list = []  # 用于查询子张量的timebin
    subgraph_timebin_idx_list = []  # 存放每个子张量时间维度在subgraph_timebin_list的下标
    subgraph_can_divided = []
    # find all suspicious edges
    for top_i in range(len(res)):
        idx1 = list(res[top_i][0][0])  # A
        idx2 = list(res[top_i][0][1])  # M
        idx3 = list(res[top_i][0][2])  # C
        time_bins = list(res[top_i][0][3])  # timebins
        subgraph_timebin_list.append(time_bins)

        susp_amt_df = amt_tensor.data.loc[amt_tensor.data[0].isin(idx1) & amt_tensor.data[1].isin(idx2) & amt_tensor.data[2].isin(time_bins)]
        susp_cmt_df = cmt_tensor.data.loc[cmt_tensor.data[0].isin(idx3) & cmt_tensor.data[1].isin(idx2) & cmt_tensor.data[2].isin(time_bins)]
        susp_amt_df.columns = ['a', 'm', 't', 'money']
        susp_cmt_df.columns = ['c', 'm', 't', 'money']
        susp_amt_df['a'] = susp_amt_df['a'].apply(lambda x: 'a' + str(x))
        susp_amt_df['m'] = susp_amt_df['m'].apply(lambda x: 'm' + str(x))
        susp_cmt_df['c'] = susp_cmt_df['c'].apply(lambda x: 'c' + str(x))
        susp_cmt_df['m'] = susp_cmt_df['m'].apply(lambda x: 'm' + str(x))

        # 构造有向图（多条边）
        G = nx.MultiDiGraph()
        G.add_weighted_edges_from(susp_amt_df[['a', 'm', 'money']].values)
        G.add_weighted_edges_from(susp_cmt_df[['m', 'c', 'money']].values)

        # 获得连通子图
        subgraph_connected_list_topi = [G.subgraph(c).copy() for c in nx.weakly_connected_components(G)]
        subgraph_connected_list.extend(subgraph_connected_list_topi)
        subgraph_timebin_idx_list.extend([top_i] * len(subgraph_connected_list_topi))

        # 维护判断是否可拆解list
        if res[top_i][1] == float("-inf") and len(subgraph_connected_list_topi) == 1:
            subgraph_can_divided.append(0) # 不可拆分

        else:
            subgraph_can_divided.extend([1] * len(subgraph_connected_list_topi))


    return subgraph_connected_list, subgraph_timebin_list, subgraph_timebin_idx_list, subgraph_can_divided

def divide_amc_dict(node_list):
    '''
    用于将str账号混合列表处理回int形式，并划分为A、M、C
    '''
    a_list = []
    m_list = []
    c_list = []
    for acc in node_list:
        acc_int = int(re.sub('[a-zA-Z]', "", acc)) # 转换为int型
        if acc.startswith('a'):
            a_list.append(acc_int)
        elif acc.startswith('m'):
            m_list.append(acc_int)
        elif acc.startswith('c'):
            c_list.append(acc_int)
        else:
            raise ValueError
    return a_list, m_list, c_list

def get_csv_from_subgraph(node_list, amt_tensor, cmt_tensor, subgraph_timebin_list, timebin_idx):
    '''
    从字符串账号列表中获得适用于算法输入的csv数据
    '''
    a_list, m_list, c_list = divide_amc_dict(node_list)
    time_bins =  subgraph_timebin_list[timebin_idx]
    am_subgraph_df = amt_tensor.data[(amt_tensor.data[0].isin(a_list)) & (amt_tensor.data[1].isin(m_list)) & (amt_tensor.data[2].isin(time_bins))]
    cm_subgraph_df = cmt_tensor.data[(cmt_tensor.data[0].isin(c_list)) & (cmt_tensor.data[1].isin(m_list)) & (cmt_tensor.data[2].isin(time_bins))]
    return am_subgraph_df, cm_subgraph_df

def relabel_acct(am_df, cm_df):
    '''
    对账号和timebin进行重新编号，加快算法运行速度
    '''
    temp_left = am_df.copy()
    temp_right = cm_df.copy()
    temp_left.columns = ['a_acct', 'm_acct', 'time_bin', 'amount']
    temp_right.columns = ['c_acct', 'm_acct', 'time_bin', 'amount']

    middle_acc_list = pd.Series(list(set(temp_left['m_acct']) & set(temp_right['m_acct']))) \
        .drop_duplicates().sort_values().reset_index(drop=True)
    temp_left = temp_left[temp_left['m_acct'].isin(middle_acc_list.values)]
    temp_right = temp_right[temp_right['m_acct'].isin(middle_acc_list.values)]

    timebin_list = temp_left['time_bin'].append(temp_right['time_bin']).drop_duplicates().sort_values().reset_index(
        drop=True)

    print('middle acct len: ', len(middle_acc_list))
    print('timebin count:', len(timebin_list))

    left_acc_list = temp_left['a_acct'].drop_duplicates().sort_values().reset_index(drop=True)  # 构建字典
    right_acc_list = temp_right['c_acct'].drop_duplicates().sort_values().reset_index(drop=True)

    # map accounts from 0
    left_acc_dict = dict(zip(list(left_acc_list), list(left_acc_list.index.astype(int))))
    middle_acc_dict = dict(zip(list(middle_acc_list), list(middle_acc_list.index.astype(int))))
    right_acc_dict = dict(zip(list(right_acc_list), list(right_acc_list.index.astype(int))))  # 构建字典完成
    timebin_dict = dict(zip(list(timebin_list), list(timebin_list.index.astype(int))))

    temp_left['a_acct'] = temp_left['a_acct'].apply(lambda x: left_acc_dict[x])  # 对原有属性进行映射
    temp_left['m_acct'] = temp_left['m_acct'].apply(lambda x: middle_acc_dict[x])
    temp_left['time_bin'] = temp_left['time_bin'].apply(lambda x: timebin_dict[x])
    temp_right['m_acct'] = temp_right['m_acct'].apply(lambda x: middle_acc_dict[x])
    temp_right['c_acct'] = temp_right['c_acct'].apply(lambda x: right_acc_dict[x])
    temp_right['time_bin'] = temp_right['time_bin'].apply(lambda x: timebin_dict[x])

    # map 0 to original account
    left_acc_dict2 = dict(zip(list(left_acc_list.index.astype(int)), list(left_acc_list)))
    middle_acc_dict2 = dict(zip(list(middle_acc_list.index.astype(int)), list(middle_acc_list)))
    right_acc_dict2 = dict(zip(list(right_acc_list.index.astype(int)), list(right_acc_list)))  # 构建字典完成
    timebin_dict2 = dict(zip(list(timebin_list.index.astype(int)), list(timebin_list)))

    temp_left.columns = temp_right.columns = [0, 1, 2, 3]

    return temp_left, temp_right, [left_acc_dict2, middle_acc_dict2, right_acc_dict2, timebin_dict2]

def call_cubeflow(am_df, cm_df, maxsize, outpath='', alpha=0.8, k=-1, dim=3, del_type=1, is_find_all_blocks=True):
    temp_left = am_df.copy()
    temp_right = cm_df.copy()

    temp_left, temp_right, dict_list = relabel_acct(temp_left, temp_right)

    amt_tensor = TensorData(temp_left)
    cmt_tensor = TensorData(temp_right)

    amt_stensor = amt_tensor.toSTensor(hasvalue=True)
    cmt_stensor = cmt_tensor.toSTensor(hasvalue=True)
    print(amt_stensor.shape)
    print(cmt_stensor.shape)

    cf = CubeFlow([amt_stensor, cmt_stensor], alpha=alpha, k=k, dim=dim, outpath=outpath)
    res = cf.run(del_type=del_type, maxsize=maxsize, is_find_all_blocks=is_find_all_blocks)

    for i in range(len(res)):
        a_list = list(map(lambda x: dict_list[0][x], res[i][0][0]))
        m_list = list(map(lambda x: dict_list[1][x], res[i][0][1]))
        c_list = list(map(lambda x: dict_list[2][x], res[i][0][2]))
        t_list = list(map(lambda x: dict_list[3][x], res[i][0][3]))
        res[i][0][0] = a_list
        res[i][0][1] = m_list
        res[i][0][2] = c_list
        res[i][0][3] = t_list

    return res

def cal_score_from_graph(subgraph, alpha):
    # 连通子图的am+cm矩阵
    subG_weight_list = list(subgraph.edges(data='weight', default=1))
    subG_weight_df = pd.DataFrame(subG_weight_list)
    print(subG_weight_df)

    subG_in_Df = subG_weight_df[subG_weight_df[1].str.contains('m')]
    subG_out_Df = subG_weight_df[subG_weight_df[0].str.contains('m')]
    subG_indegree = subG_in_Df.groupby(1).sum()
    subG_outdegree = subG_out_Df.groupby(0).sum()

    subG_indegree.index.name = 'm'
    subG_outdegree.index.name = 'm'
    subG_indegree.columns = ['indegree']
    subG_outdegree.columns = ['outdegree']

    subG_info = pd.concat([subG_indegree, subG_outdegree], axis=1, sort=False)
    subG_info['indegree'].fillna(0, inplace=True)
    subG_info['outdegree'].fillna(0, inplace=True)

    #     # 连通子图中m层节点
    #     m_layer_nodes = pd.concat([subG_in_Df[1], subG_out_Df[0]]).drop_duplicates(keep='first', inplace=False)
    #     print('m_layer_nodes len:', len(m_layer_nodes))
    #     m_node_num.append(len(m_layer_nodes))

    # 计算得分
    subG_info_array = subG_info.to_numpy()
    subG_info['f'], subG_info['q'] = np.min(subG_info_array, axis=1), np.max(subG_info_array, axis=1)
    subG_info['r'] = subG_info['q'] - subG_info['f']
    print(subG_info)

    curScore1 = subG_info['f'].sum()
    curScore2 = subG_info['r'].sum()
    s = subgraph.number_of_nodes()
    curScore = ((1 - alpha) * curScore1 - alpha * curScore2) / s
    print(f'score of g(S): {curScore}, f: {curScore1}, q-f: {curScore2}')

    return curScore

def revert_acct(amct_list, dict_dir):
    '''
    用于将a、m、c账户映射回原始字符串编号
    :param a_list: 需要映射的a账户list
    :param m_list: ...
    :param c_list: ...
    :param dict_dir: 存放映射关系文件的目录
    :return:
    '''
    a_list = amct_list[0]
    m_list = amct_list[1]
    c_list = amct_list[2]
    t_list = amct_list[3]

    dic_a = pd.read_csv(os.path.join(dict_dir, 'dict_a.csv'))
    dic_m = pd.read_csv(os.path.join(dict_dir, 'dict_m.csv'))
    dic_c = pd.read_csv(os.path.join(dict_dir, 'dict_c.csv'))
    dic_t = pd.read_csv(os.path.join(dict_dir, 'dict_time.csv'))

    dict_a = dict(zip(dic_a['new'], dic_a['ori']))  # mapped_number: origin_number
    dict_m = dict(zip(dic_m['new'], dic_m['ori']))
    dict_c = dict(zip(dic_c['new'], dic_c['ori']))
    dict_t = dict(zip(dic_t['new'], dic_t['old']))


    ori_a_list = list(map(lambda x: dict_a[x], a_list))
    ori_m_list = list(map(lambda x: dict_m[x], m_list))
    ori_c_list = list(map(lambda x: dict_c[x], c_list))
    ori_t_list = list(map(lambda x: dict_t[x], t_list))

    return [ori_a_list, ori_m_list, ori_c_list, ori_t_list]

def cal_score_from_acc(amct_list, alpha, amt_tensor, cmt_tensor):
    a_list, m_list, c_list, t_list = amct_list[0], amct_list[1], amct_list[2], amct_list[3]
    # 连通子图的am+cm矩阵
    subG_in_Df = amt_tensor.data[amt_tensor.data[0].isin(a_list) & amt_tensor.data[1].isin(m_list) & amt_tensor.data[2].isin(t_list)]
    subG_out_Df = cmt_tensor.data[cmt_tensor.data[0].isin(c_list) & cmt_tensor.data[1].isin(m_list) & cmt_tensor.data[2].isin(t_list)]
    subG_indegree = subG_in_Df[[1, 2, 3]].groupby([1, 2]).sum()
    subG_outdegree = subG_out_Df[[1, 2, 3]].groupby([1, 2]).sum()

    subG_indegree.columns = ['indegree']
    subG_outdegree.columns = ['outdegree']

    subG_info = pd.concat([subG_indegree, subG_outdegree], axis=1, sort=False)
    subG_info['indegree'].fillna(0, inplace=True)
    subG_info['outdegree'].fillna(0, inplace=True)

    print(subG_info)

    # 计算得分
    subG_info_array = subG_info.to_numpy()
    subG_info['f'], subG_info['q'] = np.min(subG_info_array, axis=1), np.max(subG_info_array, axis=1)
    subG_info['r'] = subG_info['q'] - subG_info['f']
    print(subG_info)

    curScore1 = subG_info['f'].sum()
    curScore2 = subG_info['r'].sum()

    time_df = subG_in_Df[2].append(subG_out_Df[2]).drop_duplicates().sort_values().reset_index(drop=True)
    t_len = time_df.shape[0]

    s = t_len + len(m_list) + len(a_list) + len(c_list)
    curScore = ((1 - alpha) * curScore1 - alpha * curScore2) / s
    print(f'score of g(S): {curScore}, f: {curScore1}, q-f: {curScore2}, s:{s}')

    return curScore, curScore1, curScore2

def cal_score_from_acc_sparse(amct_list, alpha, csr_matrix_list, timebin2idx_dict):
    a_list, m_list, c_list, t_list = amct_list[0], amct_list[1], amct_list[2], amct_list[3]
    t_len, curScore1, curScore2 = 0, 0, 0
    for t in t_list:
        am_matrix = csr_matrix_list[timebin2idx_dict[t]][0]
        am_matrix_sub = am_matrix[a_list, :].tocsc()[:, m_list]
        cm_matrix = csr_matrix_list[timebin2idx_dict[t]][1]
        cm_matrix_sub = cm_matrix[c_list, :].tocsc()[:, m_list]

        if am_matrix_sub.nnz == 0 and cm_matrix_sub.nnz == 0:
            continue
        else:
            t_len += 1

        indegree_arr = np.squeeze(am_matrix_sub.sum(axis=0).A)
        outdegree_arr = np.squeeze(cm_matrix_sub.sum(axis=0).A)

        if indegree_arr.shape == ():
            indegree_arr = np.array([indegree_arr.tolist()])
            outdegree_arr = np.array([outdegree_arr.tolist()])

        mid_min = []
        for m1, m2 in zip(indegree_arr, outdegree_arr):
            mid_min.append(min(m1, m2))  # fi

        curScore1 += sum(mid_min)
        curScore2 += sum(abs(indegree_arr - outdegree_arr))

    s = t_len + len(m_list) + len(a_list) + len(c_list)
    curScore = ((1 - alpha) * curScore1 - alpha * curScore2) / s
    print(f'score of g(S): {curScore}, f: {curScore1}, q-f: {curScore2}, s:{s}')
    return curScore, curScore1, curScore2

def cal_score_from_acc_sparse_fs(amct_list, alpha, a_mt_mat, c_mt_mat):
    a_list, m_list, c_list, t_list = amct_list[0], amct_list[1], amct_list[2], amct_list[3]

    a_mt_mat_sub = a_mt_mat[a_list, :].tocsc()[:, m_list]
    c_mt_mat_sub = c_mt_mat[c_list, :].tocsc()[:, m_list]

    indegree_arr = np.squeeze(a_mt_mat_sub.sum(axis=0).A)
    outdegree_arr = np.squeeze(c_mt_mat_sub.sum(axis=0).A)

    if indegree_arr.shape == ():
        indegree_arr = np.array([indegree_arr.tolist()])
        outdegree_arr = np.array([outdegree_arr.tolist()])

    mid_min = []
    for m1, m2 in zip(indegree_arr, outdegree_arr):
        mid_min.append(min(m1, m2))  # fi

    curScore1, curScore2 = 0, 0

    curScore1 += sum(mid_min)
    curScore2 += sum(abs(indegree_arr - outdegree_arr))

    s = len(m_list) + len(a_list) + len(c_list)
    curScore = ((1 - alpha) * curScore1 - alpha * curScore2) / s
    # print(f'score of g(S): {curScore}, f: {curScore1}, q-f: {curScore2}, s:{s}')
    return curScore, curScore1, curScore2

def get_zero_matrix(idx, size):
    row = np.array(list(range(size)))
    col = np.array(list(range(size)))
    data = np.array([1]*size)
    row = np.delete(row, idx)
    col = np.delete(col, idx)
    data = np.delete(data, idx)
    return ssp.csr_matrix((data, (row, col)), shape=(size, size))

def loadtxt2res(outpath, dim, data_type=int):
    res = []
    tmp_res = []
    with open(outpath, "r") as f:
        contents =f.readlines()
        for line in contents:
            if line.startswith('['):
                tmp_res = []
                continue
            if line.startswith(']'):
                res.append(tmp_res)
                continue
            acc_list = line.strip().split("\t")
            if len(tmp_res) < dim + 1:
                acc_list = list(map(data_type, acc_list))
                tmp_res.append(acc_list)
            else:
                score = float(acc_list[0])
                tmp_res = [tmp_res, score]
    print('load res.txt success!')
    return res

def handle_big_graph(subgraph_connected_list, subgraph_timebin_list, subgraph_timebin_idx_list, subgraph_can_divided, max_node_limit,
                     amt_tensor, cmt_tensor, handle_biggraph_type, maxsize2, alpha, dim, del_type):
    # 步骤二：对每个大图（总节点数超过最大限制）进行单独处理
    # 选择1：去掉度最大的节点，放入连通图算法
    # 选择2：运行带约束的CubeFlow，直至跑空所有节点为止
    curr_i = 0
    while curr_i < len(subgraph_connected_list):
        curr_subgraph = subgraph_connected_list[curr_i]
        curr_nodes_list = list(curr_subgraph.nodes())
        a_list_tmp, m_list_tmp, c_list_tmp = divide_amc_dict(curr_nodes_list)
        if len(m_list_tmp) > max_node_limit:

            if handle_biggraph_type == 1:
                m_str_list_tmp = ['m' + str(m) for m in m_list_tmp]
                curr_degree_dict = dict(curr_subgraph.degree(m_str_list_tmp))  # 去掉m账户中度最大的节点

                max_degree_node = max(curr_degree_dict, key=curr_degree_dict.get)
                curr_matrix = nx.to_scipy_sparse_matrix(curr_subgraph, curr_nodes_list)

                node_idx = curr_nodes_list.index(max_degree_node)

                zero_matrix = get_zero_matrix(node_idx, curr_matrix.shape[0])
                curr_matrix = zero_matrix * curr_matrix * zero_matrix  # 删除结点
                curr_subgraph = nx.from_scipy_sparse_matrix(curr_matrix,
                                                            create_using=nx.MultiDiGraph)

                mapping = {}
                for i in range(len(curr_nodes_list)):
                    mapping[i] = curr_nodes_list[i]

                curr_subgraph = nx.relabel_nodes(curr_subgraph, mapping)
                curr_subgraph.remove_node(max_degree_node)
                print(f'max degree node [{max_degree_node}] has been removed!')

                # 重新运行连通子图算法
                curr_subgraph_connected_list = [curr_subgraph.subgraph(c).copy() for c in
                                                nx.weakly_connected_components(curr_subgraph)]
                print('add subgraph num:', len(curr_subgraph_connected_list))

                subgraph_connected_list.extend(curr_subgraph_connected_list)
                tb_idx = subgraph_timebin_idx_list[curr_i]
                subgraph_timebin_idx_list.append(tb_idx)

            elif handle_biggraph_type == 2:

                if subgraph_can_divided[curr_i] == 0:
                    curr_i += 1
                    continue

                # 准备算法输入数据
                am_df, cm_df = get_csv_from_subgraph(curr_nodes_list, amt_tensor, cmt_tensor, subgraph_timebin_list,
                                                     subgraph_timebin_idx_list[curr_i])

                # 调用CubeFlow算法
                curr_res = call_cubeflow(am_df, cm_df, maxsize=maxsize2, outpath='', alpha=alpha, k=-1, dim=dim,
                                         del_type=del_type, is_find_all_blocks=True)
                print('find subgraph num:', len(curr_res))

                # 复用拆分连通子图代码
                curr_subgraph_connected_list, curr_subgraph_timebin_list, curr_subgraph_timebin_idx_list, curr_subgraph_can_divided \
                    = divide_connected_conponents(curr_res, amt_tensor, cmt_tensor)
                print('add subgraph num:', len(curr_subgraph_connected_list))

                # 加入现有的连通子图list中
                subgraph_connected_list.extend(curr_subgraph_connected_list)
                old_timebin_list_len = len(subgraph_timebin_list)
                curr_subgraph_timebin_idx_list = list(
                    map(lambda x: x + old_timebin_list_len, curr_subgraph_timebin_idx_list))

                subgraph_timebin_idx_list += curr_subgraph_timebin_idx_list
                subgraph_timebin_list += curr_subgraph_timebin_list

                subgraph_can_divided += curr_subgraph_can_divided

                del subgraph_can_divided[curr_i]

            else:
                raise ValueError

            del subgraph_timebin_idx_list[curr_i]
            del subgraph_connected_list[curr_i]


        else:
            curr_i += 1

    return subgraph_connected_list, subgraph_timebin_list, subgraph_timebin_idx_list

def sparse_df_to_csr(df, matrix_shape):
    tensor = TensorData(df)
    stensor = tensor.toSTensor(hasvalue=True)
    tmp = stensor._data
    tmp = ssp.coo_matrix((tmp.data, (tmp.coords[0], tmp.coords[1])), shape=matrix_shape)
    return tmp.tocsr()

def prepare_for_sparse_score(subgraph_timebin_list, amt_stensor, cmt_stensor, amt_tensor, cmt_tensor):
    subgraph_timebin_list_flatten = []
    for i in range(len(subgraph_timebin_list)):
        subgraph_timebin_list_flatten.extend(subgraph_timebin_list[i])
    subgraph_timebin_list_flatten = list(set(subgraph_timebin_list_flatten))

    am_matrix_shape = amt_stensor.shape[:2]
    cm_matrix_shape = cmt_stensor.shape[:2]
    print(am_matrix_shape, cm_matrix_shape)

    csr_matrix_list = []
    for i in range(len(subgraph_timebin_list_flatten)):
        am_df = amt_tensor.data[amt_tensor.data[2] == subgraph_timebin_list_flatten[i]][[0, 1, 3]]
        cm_df = cmt_tensor.data[cmt_tensor.data[2] == subgraph_timebin_list_flatten[i]][[0, 1, 3]]

        am_matrix = sparse_df_to_csr(am_df, am_matrix_shape)
        cm_matrix = sparse_df_to_csr(cm_df, cm_matrix_shape)
        csr_matrix_list.append([am_matrix, cm_matrix])

    timebin2idx_dict = {}
    for i in range(len(subgraph_timebin_list_flatten)):
        timebin2idx_dict[subgraph_timebin_list_flatten[i]] = i

    return csr_matrix_list, timebin2idx_dict

def score_connected_graph_list(subgraph_connected_list, subgraph_timebin_list, subgraph_timebin_idx_list,
                               alpha, amt_stensor, cmt_stensor, amt_tensor, cmt_tensor):
    print('begin to score connected subgraph....')

    # csr_matrix_list, timebin2idx_dict = prepare_for_sparse_score(subgraph_timebin_list, amt_stensor, cmt_stensor,
    #                                                              amt_tensor, cmt_tensor)

    # 算法函数使用fs目标函数，即不包含t时使用
    a_mt_mat = amt_stensor.sum(axis=2)._data.tocsr()
    c_mt_mat = cmt_stensor.sum(axis=2)._data.tocsr()

    subgraph_score_list = []
    subgraph_score1_list = []  # f
    subgraph_score2_list = []  # q - f
    amct_list = []
    total_len = len(subgraph_connected_list)

    for idx in range(len(subgraph_connected_list)):
        # print(f'subgraph {idx}/{total_len}: ', end='')
        subgraph = subgraph_connected_list[idx]
        a_list_tmp, m_list_tmp, c_list_tmp = divide_amc_dict(list(subgraph.nodes()))

        amct_list_sub = [a_list_tmp, m_list_tmp, c_list_tmp, subgraph_timebin_list[subgraph_timebin_idx_list[idx]]]
        # curScore, curScore1, curScore2 = cal_score_from_acc_sparse(amct_list_sub, alpha, csr_matrix_list,
        #                                                            timebin2idx_dict)

        curScore, curScore1, curScore2 = cal_score_from_acc_sparse_fs(amct_list_sub, alpha, a_mt_mat, c_mt_mat)

        subgraph_score_list.append(curScore)
        subgraph_score1_list.append(curScore1)
        subgraph_score2_list.append(curScore2)
        amct_list.append(amct_list_sub)

    return subgraph_score_list, subgraph_score1_list, subgraph_score2_list, amct_list

def mkdir_self(dir):
    if dir.endswith('.txt'):
        dir_list = dir.split('/')
        dir = '/'.join(dir_list[:-1])

    if not os.path.exists(dir):
        os.makedirs(dir)

def saveres2txt(res, outpath):
    mkdir_self(outpath)
    for i in range(len(res[0])):
        res[0][i] = list(res[0][i])
    with open(outpath, "a") as f:
        f.write('[')
        f.write('\n')
        for i in range(len(res[0])):
            for j in range(len(res[0][i])):
                f.write(str(res[0][i][j]))
                f.write('\t')
            f.write('\n')
        f.write(str(res[1]))
        f.write('\n')
        f.write(']')
        f.write('\n')
    # print('save res.txt success!')

def loadlisttxt2res(outpath):
    res = []
    with open(outpath, "r") as f:
        contents =f.readlines()
        for line in contents:
            res.append(str(line.strip()))
    print('load res.txt success!')
    return res



try:
    from . import c_MDL
except Exception as e:
    import os
    old_dir = os.getcwd()
    dir_ = os.path.abspath(os.path.dirname(__file__))
    os.chdir(dir_)
    import pyximport
    pyximport.install(build_in_temp=False, inplace=True)
    os.chdir(old_dir)
    from . import c_MDL

from .DPGS import DPGSummarizer

def __call__():
    return DPGSummarizer


#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

# #  Statistic hypothesis test class
#  Author: wenchieh
#
#  Project: eaglemine
#      statistic_hypothesis_test.py
#      Version:  1.0
#      Date: December 05 2017
#      Main Contact: Wenchieh Feng (wenchiehfeng.us@gmail.com)
#
#      Copyright:
#        This software is free of charge under research purposes.
#        For commercial purposes, please contact the author.
#
#      Created by @wenchieh  on <12/5/2017>
#

__author__ = 'wenchieh'

# third-party lib
import numpy as np
from numpy.linalg import eig
from sklearn.preprocessing import scale
from scipy.stats import anderson

# project
from .norm_extras import norm_anderson


class StatisticHypothesisTest(object):
    def __init__(self, strictness=4):
        self.strict = strictness
        self._min_float = np.finfo('float64').resolution

    def _expand_(self, xs, count):
        if np.max(count) == 1:
            return np.array(xs)
        else:
            exp_xs = list()
            for k in range(len(xs)):
                exp_xs.extend([list(xs[k])] * count[k])
            return np.asarray(exp_xs)

    def _gaussian_anderson_(self, vector, strictness):
        """
        check whether a given input vector follows a gaussian distribution
        H0: vector is distributed gaussian
        H1: vector is not distributed gaussian
        :param vector [array], all_data to be checked
        :param strictness [int default=4], {0, 1, 2, 3, 4}
                     how strict should the anderson-darling test for normality be
                     0: not at all strict;     4: very strict
             Critical values provided are for the following significance levels:
               normal/exponential:     15%, 10%, 5%, 2.5%, 1%
        """
        output = anderson(vector)
        if output[0] <= output[1][strictness]:
            return True
        else:
            return False

    def _gaussian_check(self, xs ,cov):
        """
        project data [xs] into eigenvectors specific axises to be one-dimensional for statistical test
        References
        ----------
         [1] Hamerly, Greg and Charles Elkan. “Learning the k in k-means.” NIPS (2003).
        ----------
        """
        _, v = eig(np.asarray(cov))
        gaus = True
        for k in range(len(cov)):
            kv = v[:, k]
            proj_kvs = scale(xs.dot(kv) / (kv.dot(kv)) + self._min_float)
            gaus &= self._gaussian_anderson_(proj_kvs, self.strict)
            if not gaus:
                break

        return gaus

    def _mixture_assign_scale_(self, xs, mus, covs, scale=5):
        assigns = list()
        nxs = len(xs)
        components, dims = np.array(covs).shape

        for i in range(components):
            mu, cov = mus[i], covs[i]
            _, v = eig(np.asarray(cov))
            belongs = np.array([True] * nxs)
            for k in range(dims):
                kv = v[:, k]
                sigma = np.sqrt(cov[k, k])
                proj_kv = abs((xs - mu).dot(kv) / (kv.dot(kv)))
                belongs &= (proj_kv <= scale * sigma)
            assigns.append(xs[belongs, :])

    def _mixture_assign_prob_(self, xs, xs_probs, components, mix_weights):
        assigns = list()
        xs_wprob = np.array(xs_probs) * np.array(mix_weights)
        xs_comp_ = np.argmax(xs_wprob, axis=1)
        for k in range(components):
            assigns.append(np.where(xs_comp_ == k)[0])

        return assigns

    def apply(self, xs, xs_weights, xs_probs,  paras, is_mix, argv=None):
        xs = np.asarray(xs)
        xs = self._expand_(xs, np.array(np.ceil(xs_weights), int))

        if is_mix:
            mus, covs, mix_wt = paras.get("mus"), paras.get("covs"), paras.get("weights")
            n_mixs = len(mus)
            xs_probs = self._expand_(xs_probs, np.array(np.ceil(xs_weights), int))
            assigns = self._mixture_assign_prob_(xs, xs_probs, n_mixs, mix_wt)
            ## assigns = self._mixture_assign_scale_(xs, mus, covs, scale=5)
            mix_gaus = True
            for i in range(n_mixs):
                xs_imix = xs[assigns[i], :]
                mix_gaus &= self._gaussian_check(xs_imix, covs[i])
                if not mix_gaus:
                    break
            return mix_gaus
        else:
            cov = paras.get("covs")[0]
            return self._gaussian_check(xs, cov)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   anomaly_detection.py
@Desc    :   Implementation for anomaly detection task.
'''

# here put the import lib

from . import MODEL_PATH

from ._task import Task
from enum import Enum


class AnomalyDetection(Task):
    '''Implementation for anomaly detection task.
    '''

    def run(self, *args, **kwargs):
        '''Call anomaly detection function of selected model.

        If not implemented, raise an exception by calling parent run.
        '''
        if "anomaly_detection" in dir(self.model):
            return self.model.anomaly_detection(*args, **kwargs)
        else:
            return super().run("anomaly_detection")

    @classmethod
    def list_models(cls):
        '''List models which can do anomaly detection task.
        '''
        for name in ADPolicy.__members__:
            print(name)


class ADPolicy(Enum):
    '''Registration for path of models who can do anomaly detection task.
    '''
    # EagleMine = MODEL_PATH + ".eaglemine"
    EigenPulse = MODEL_PATH + ".eigenpulse"
    HoloScope = MODEL_PATH + ".holoscope"
    FlowScope = MODEL_PATH + ".flowscope"
    pass



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   basicutil.py
@Desc    :   Basic util functions.
'''

# here put the import lib
import time
import numpy as np
from datetime import datetime
from bisect import bisect_left


def param_default(params: dict, key: str, default):
    if type(default) is dict:
        return params[key] if params.__contains__(key) else default[key]
    else:
        return params[key] if params.__contains__(key) else default


def set_trace(isset=True):
    if isset is True:
        import ipdb
        ipdb.set_trace()
    else:
        pass


class _Mapper:
    def __init__(self):
        pass

    def map(self, attrs):
        pass

    def revert(self, indices):
        pass


class TimeMapper(_Mapper):
    def __init__(self, timeformat='%Y-%m-%d', timebin=24*3600, mints=None):
        self.timeformat = timeformat
        self.timebin = timebin
        self.mints = mints

    def map(self, attrs):
        tindices = []
        set_trace(False)
        for t in attrs:
            da = datetime.strptime(t, self.timeformat)
            ts = time.mktime(da.timetuple())
            tindices.append(int(ts))

        tindices = np.array(tindices)
        if self.mints is None:
            self.mints = tindices.min()

        tindices = (tindices - self.mints) // self.timebin
        return tindices

    def revert(self, indices):

        timestr = []
        tss = indices + self.mints + self.timebin//2
        for ts in tss:
            date_time = datetime.fromtimestamp(ts)
            d = date_time.strftime(self.timeformat)
            timestr.append(d)

        return np.array(timestr)


class ScoreMapper(_Mapper):
    ''' map score (int or float) into index (int)

    Parameters
    ----------
    scorebin : If scorebin is int, evenly divide from min(score) to max(score)
    into scorebin segments. If scorebin is a list, map score with list values
    as bounds.
    [0, 2.0, 3.0]

    '''

    def __init__(self, scorebin):
        self.scorebin = scorebin
        self.residules = None

    def map(self, attrs):
        sindices = []
        residules = []
        scorebin = self.scorebin
        if isinstance(scorebin, int):
            scorebin = [attrs.min(), scorebin, attrs.max()]
        elif isinstance(scorebin, list):
            _min, _max = min(attrs), max(attrs)
            if _min not in scorebin:
                scorebin.append(_min)
            if _max not in scorebin:
                scorebin.append(_max)
            scorebin.sort()
        for score in attrs:
            if score in scorebin:
                sindices.append(scorebin.index(score))
                residules.append(0)
            else:
                pos = bisect_left(scorebin, score)
                if scorebin[pos] > score:
                    pos -= 1
                    sindices.append(pos)
                else:
                    sindices.append(pos)
                residules.append(score-scorebin[pos])
        self.residules = residules
        self.scorebin = scorebin
        return sindices

    def revert(self, indices):
        residules = self.residules
        scorebin = self.scorebin
        scores = []
        for ind, val in enumerate(indices):
            scores.append(scorebin[val] + residules[ind])
        return scores


class StringMapper(_Mapper):
    '''mapping the names or complex string ids of users and objects into
    indices. Note that if the matrix is homogenous, i.e. user-to-user
    relations, then use the same StringMapper instance for both
    user-to-user columns. The two mappers will share the same strdict, so
    mapping their ids into the sample space.
    '''

    def __init__(self, strdict: dict = {}):
        self.strdict = strdict
        self.strids = []
        if len(self.strdict) >= 0:
            self._append_string_ids()
        pass

    def _append_string_ids(self):
        for k, v in self.strdict.items():
            self.strids.append(k)

    def map(self, attrs):
        indices = {}
        for s in attrs:
            if s not in self.strdict:
                self.strdict[s] = len(self.strdict)
                self.strids.append(s)  # add new ids
            indices = self.strdict[s]

        return indices

    def revert(self, indices):
        attrs = []

        for i in indices:
            attrs.append(self.strids[i])

        return attrs


class IntMapper(_Mapper):
    '''Shift the ints starting from zero
    '''

    def __init__(self, minint=None):
        self.minint = minint
        pass

    def map(self, attrs):
        if self.minint is not None:
            self.minint = min(attrs)
        indices = attrs-self.minint
        return indices

    def revert(self, indices):
        attrs = indices + self.minint
        return attrs


class DenseIntMapper(_Mapper):
    """Remapp irregular integers starting from zero
    """

    def __init__(self):
        self.id2int = dict()
        self.int2id = dict()
        self._idx = 0

    def map(self, attrs):
        rets = [None] * len(attrs)
        for i, attr in enumerate(attrs):
            if not (attr in self.int2id):
                self.int2id[attr] = self._idx
                self.id2int[self._idx] = attr
                self._idx += 1
            rets[i] = self.int2id[attr]
        return rets

    def revert(self, indices):
        return [self.id2int[idx] for idx in indices]
